r"""
    mermaid_inheritance
    ~~~~~~~~~~~~~~~~~~~

    Modified by the CoSApp team from sphinx.ext.inheritance_diagram

    Defines a docutils directive for inserting inheritance diagrams.
    Provide the directive with one or more classes or modules (separated
    by whitespace).  For modules, all of the classes in that module will
    be used.
    Example::
       Given the following classes:
       class A: pass
       class B(A): pass
       class C(A): pass
       class D(B, C): pass
       class E(B): pass
       .. inheritance-diagram: D E
       Produces a graph like the following:
                   A
                  / \
                 B   C
                / \ /
               E   D
    The graph is inserted as a PNG+image map into HTML and a PDF in
    LaTeX.
    :copyright: Copyright 2007-2020 by the Sphinx team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import builtins
import inspect
import re
from hashlib import md5
from importlib import import_module
from pathlib import Path
from typing import Any, Dict, Iterable, List, Tuple, cast

import sphinx
from docutils import nodes
from docutils.nodes import Node
from docutils.parsers.rst import directives
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.environment import BuildEnvironment
from sphinx.ext.inheritance_diagram import (
    InheritanceDiagram,
    InheritanceException,
    InheritanceGraph,
    figure_wrapper,
    get_graph_hash,
    inheritance_diagram,
    skip,
)
from sphinx.util.docutils import SphinxDirective
from sphinx.writers.html import HTMLTranslator
from sphinx.writers.latex import LaTeXTranslator
from sphinx.writers.texinfo import TexinfoTranslator

from mermaid import render_mm_html, render_mm_latex, render_mm_texinfo


class MermaidGraph(InheritanceGraph):
    """
    Given a list of classes, determines the set of classes that they inherit
    from all the way to the root "object", and then is able to generate a
    mermaid graph from them.
    """

    # These are the default attrs
    default_graph_attrs = {}
    #     'rankdir': 'LR',
    #     'size': '"8.0, 12.0"',
    #     'bgcolor': 'transparent',
    # }
    default_node_attrs = {}
    #     'shape': 'box',
    #     'fontsize': 10,
    #     'height': 0.25,
    #     'fontname': '"Vera Sans, DejaVu Sans, Liberation Sans, '
    #                 'Arial, Helvetica, sans"',
    #     'style': '"setlinewidth(0.5),filled"',
    #     'fillcolor': 'white',
    # }
    default_edge_attrs = {}
    #     'arrowsize': 0.5,
    #     'style': '"setlinewidth(0.5)"',
    # }

    def _format_node_attrs(self, attrs: Dict) -> str:
        # return ','.join(['%s=%s' % x for x in sorted(attrs.items())])
        return ""

    def _format_graph_attrs(self, attrs: Dict) -> str:
        # return ''.join(['%s=%s;\n' % x for x in sorted(attrs.items())])
        return ""

    def generate_dot(
        self,
        name: str,
        urls: Dict = {},
        env: BuildEnvironment = None,
        graph_attrs: Dict = {},
        node_attrs: Dict = {},
        edge_attrs: Dict = {},
    ) -> str:
        """Generate a mermaid graph from the classes that were passed in
        to __init__.
        *name* is the name of the graph.
        *urls* is a dictionary mapping class names to HTTP URLs.
        *graph_attrs*, *node_attrs*, *edge_attrs* are dictionaries containing
        key/value pairs to pass on as graphviz properties.
        """
        # g_attrs = self.default_graph_attrs.copy()
        # n_attrs = self.default_node_attrs.copy()
        # e_attrs = self.default_edge_attrs.copy()
        # g_attrs.update(graph_attrs)
        # n_attrs.update(node_attrs)
        # e_attrs.update(edge_attrs)
        # if env:
        #     g_attrs.update(env.config.inheritance_graph_attrs)
        #     n_attrs.update(env.config.inheritance_node_attrs)
        #     e_attrs.update(env.config.inheritance_edge_attrs)

        res = []  # type: List[str]

        res.append("classDiagram\n")
        for name, fullname, bases, tooltip in sorted(self.class_info):
            # Write the node
            res.append("  class {!s}\n".format(name))
            if fullname in urls:
                res.append(
                    '  link {!s} "./{!s}" {!s}\n'.format(
                        name, urls[fullname], tooltip or '"{}"'.format(name)
                    )
                )

            # Write the edges
            for base_name in bases:
                res.append("  {!s} <|-- {!s}\n".format(base_name, name))

        return "".join(res)


class mermaid_inheritance(inheritance_diagram):
    """
    A docutils node to use as a placeholder for the inheritance diagram.
    """
    pass


class MermaidDiagram(InheritanceDiagram):
    """
    Run when the mermaid_inheritance directive is first encountered.
    """
    has_content = False
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {
        "parts": int,
        "private-bases": directives.flag,
        "caption": directives.unchanged,
        "top-classes": directives.unchanged_required,
    }

    def run(self) -> List[Node]:
        node = mermaid_inheritance()
        node.document = self.state.document
        class_names = self.arguments[0].split()
        class_role = self.env.get_domain("py").role("class")
        # Store the original content for use as a hash
        node["parts"] = self.options.get("parts", 0)
        node["content"] = ", ".join(class_names)
        node["top-classes"] = []
        for cls in self.options.get("top-classes", "").split(","):
            cls = cls.strip()
            if cls:
                node["top-classes"].append(cls)

        # Create a graph starting with the list of classes
        try:
            graph = MermaidGraph(
                class_names,
                self.env.ref_context.get("py:module"),
                parts=node["parts"],
                private_bases="private-bases" in self.options,
                aliases=self.config.inheritance_alias,
                top_classes=node["top-classes"],
            )
        except InheritanceException as err:
            return [node.document.reporter.warning(err, line=self.lineno)]

        # Create xref nodes for each target of the graph's image map and
        # add them to the doc tree so that Sphinx can resolve the
        # references to real URLs later.  These nodes will eventually be
        # removed from the doctree after we're done with them.
        for name in graph.get_all_class_names():
            refnodes, x = class_role(  # type: ignore
                "class", ":class:`%s`" % name, name, 0, self.state
            )  # type: ignore
            node.extend(refnodes)
        # Store the graph object so we can use it to generate the
        # dot file later
        node["graph"] = graph

        if "caption" not in self.options:
            self.add_name(node)
            return [node]
        else:
            figure = figure_wrapper(self, node, self.options["caption"])
            self.add_name(figure)
            return [figure]


def html_visit_mermaid_inheritance(self: HTMLTranslator, node: inheritance_diagram) -> None:
    """
    Output the graph for HTML.  This will insert a PNG with clickable
    image map.
    """
    graph = node['graph']

    graph_hash = get_graph_hash(node)
    name = 'inheritance%s' % graph_hash

    # Create a mapping from fully-qualified class names to URLs.
    mermaid_output_format = self.builder.env.config.mermaid_output_format.upper()
    current_filename = self.builder.current_docname + self.builder.out_suffix
    urls = {}
    pending_xrefs = cast(Iterable[addnodes.pending_xref], node)
    for child in pending_xrefs:
        if child.get('refuri') is not None:
            if mermaid_output_format == 'SVG':
                urls[child['reftitle']] = "../" + child.get('refuri')
            else:
                urls[child['reftitle']] = child.get('refuri')
        elif child.get('refid') is not None:
            if mermaid_output_format == 'SVG':
                urls[child['reftitle']] = '../' + current_filename + '#' + child.get('refid')
            else:
                urls[child['reftitle']] = '#' + child.get('refid')
    dotcode = graph.generate_dot(name, urls, env=self.builder.env)
    render_mm_html(self, node, dotcode, {}, 'inheritance', 'inheritance',
                    alt='Inheritance diagram of ' + node['content'])
    raise nodes.SkipNode


def latex_visit_mermaid_inheritance(self: LaTeXTranslator, node: inheritance_diagram) -> None:
    """
    Output the graph for LaTeX.  This will insert a PDF.
    """
    graph = node['graph']

    graph_hash = get_graph_hash(node)
    name = 'inheritance%s' % graph_hash

    dotcode = graph.generate_dot(name, env=self.builder.env, )
                                #  graph_attrs={'size': '"6.0,6.0"'})
    render_mm_latex(self, node, dotcode, {}, 'inheritance')
    raise nodes.SkipNode


def texinfo_visit_mermaid_inheritance(self: TexinfoTranslator, node: inheritance_diagram) -> None:
    """
    Output the graph for Texinfo.  This will insert a PNG.
    """
    graph = node['graph']

    graph_hash = get_graph_hash(node)
    name = 'inheritance%s' % graph_hash

    dotcode = graph.generate_dot(name, env=self.builder.env, )
                                #  graph_attrs={'size': '"6.0,6.0"'})
    render_mm_texinfo(self, node, dotcode, {}, 'inheritance')
    raise nodes.SkipNode


def setup(app: Sphinx) -> Dict[str, Any]:
    app.setup_extension("mermaid")
    app.add_node(
        mermaid_inheritance,
        latex=(latex_visit_mermaid_inheritance, None),
        html=(html_visit_mermaid_inheritance, None),
        text=(skip, None),
        man=(skip, None),
        texinfo=(texinfo_visit_mermaid_inheritance, None),
    )
    app.add_directive("mermaid-inheritance", MermaidDiagram)
    # app.add_config_value('mermaid_inheritance_graph_attrs', {}, False)
    # app.add_config_value('mermaid_inheritance_node_attrs', {}, False)
    # app.add_config_value('mermaid_inheritance_edge_attrs', {}, False)
    app.add_config_value("inheritance_alias", {}, False)

    return {"version": sphinx.__display_version__, "parallel_read_safe": True}



import os
import json

from typing import Any, Dict, Union

from collections import OrderedDict
import logging
import base64

from cosapp.systems import System
from cosapp.ports.port import ExtensiblePort
from cosapp.ports.enum import PortType

from cosapp.utils.helpers import check_arg

logger = logging.getLogger(__name__)


def _get_tree_dict(
    system_or_port: Union[System, ExtensiblePort],
    include_wards: bool
) -> Dict[str, Any]:
    """Get a dictionary representation of the system hierarchy."""
    check_arg(system_or_port, 'system_or_port', (System, ExtensiblePort))

    tree_dict = OrderedDict(name=system_or_port.name, type='subsystem')
    children = list()

    if isinstance(system_or_port, ExtensiblePort):  # look for variable in in the system

        tree_dict['subsystem_type'] = 'component'
        if system_or_port.direction == PortType.IN:
            view = lambda n: OrderedDict(name=n, type='param')
        else:  # system_or_port.direction == PortType.OUT
            view = lambda n: OrderedDict(name=n, type='unknown')

        if include_wards or system_or_port.name not in (System.INWARDS, System.OUTWARDS):
            children.extend(map(view, iter(system_or_port)))
    
    else:

        tree_dict['subsystem_type'] = 'group'
        for port in filter(lambda p: len(p) > 0, system_or_port.inputs.values()):
            children.append(_get_tree_dict(port, include_wards))
            
        for child in system_or_port.exec_order:
            children.append(
                _get_tree_dict(
                    system_or_port.children[child], 
                    include_wards
                )
            )

        # outputs come after child for a matter of visualization
        for port in filter(lambda p: len(p) > 0, system_or_port.outputs.values()):
            children.append(_get_tree_dict(port, include_wards))
    
    tree_dict['children'] = children

    return tree_dict


def _get_connections(
    system: System,
    include_wards: bool
) -> Dict[str, str]:
    """Get a dictionary representation of the system connections. Structure is {'_in': '_out'}."""
    connections = OrderedDict()
    def add_prefix(name, s):
        if s.startswith(name):  # Case of port of systems group connect inside and outside
            return s
        else:
            return f"{name}.{s}"

    # Gather the connections for the current system
    for c in system.connectors.values():
        if c.sink.name in (System.INWARDS, System.OUTWARDS) and not include_wards:
            continue

        for target, origin in c.variable_mapping.items():
            connections[f"{c.sink.contextual_name}.{target}"] = f"{c.source.contextual_name}.{origin}"

    # Recursively gather children connections
    for name, child in system.children.items():
        child_connections = _get_connections(child, include_wards)
        for k, v in child_connections.items():
            connections[add_prefix(name, k)] = add_prefix(name, v)

    return connections

def _get_viewer_data(system: System,
                     include_wards: bool):
    """Get the data needed by the N2 viewer as a dictionary."""
    if isinstance(system, System):
        root_group = system
    else:
        raise TypeError('get_model_viewer_data only accepts System')

    if not include_wards:
        logger.warning('The system may contain inwards or outwards.')

    data_dict = {}
    data_dict['tree'] = _get_tree_dict(root_group, include_wards)
    connections_list = []
    for in_abs, out_abs in _get_connections(root_group, include_wards).items():
        if out_abs is None:
            continue
        connections_list.append(OrderedDict([('src', out_abs), ('tgt', in_abs)]))
    data_dict['connections_list'] = connections_list

    return data_dict


def view_model(problem_or_filename,
               outfile='n2.html',
               show_browser=True,
               embeddable=False,
               draw_potential_connections=False,
               include_wards=True):
    """
    Generates an HTML file containing a tree viewer. Optionally pops up a web browser to
    view the file.

    Parameters
    ----------
    problem_or_filename : A System
        System : The System for the desired tree.

    outfile : str, optional
        The name of the final output file

    show_browser : bool, optional
        If True, pop up the system default web browser to view the generated html file.
        Defaults to True.

    embeddable : bool, optional
        If True, gives a single HTML file that doesn't have the <html>, <DOCTYPE>, <body>
        and <head> tags. If False, gives a single, standalone HTML file for viewing.

    draw_potential_connections : bool, optional
        If true, allows connections to be drawn on the N2 that do not currently exist
        in the model. Defaults to True.

    include_wards : bool, optional
    If True, display inwards and outwards on the N2 diagram. Defaults to True.
    """
    html_begin_tags = """<!DOCTYPE html>
<html>
    <head>
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
    </head>
    <body>\n"""
    html_end_tags = """
    </body>
</html>"""

    code_dir = os.path.dirname(os.path.abspath(__file__))
    vis_dir = os.path.join(code_dir, "visualization")
    libs_dir = os.path.join(vis_dir, "libs")
    src_dir = os.path.join(vis_dir, "src")
    style_dir = os.path.join(vis_dir, "style")

    #grab the libraries
    with open(os.path.join(libs_dir, "awesomplete.js"), "r") as f:
        awesomplete = f.read()
    with open(os.path.join(libs_dir, "d3.v4.min.js"), "r") as f:
        d3 = f.read()
    with open(os.path.join(libs_dir, "vkBeautify.js"), "r") as f:
        vk_beautify = f.read()

    #grab the src
    with open(os.path.join(src_dir, "constants.js"), "r") as f:
        constants = f.read()
    with open(os.path.join(src_dir, "draw.js"), "r") as f:
        draw = f.read()
    with open(os.path.join(src_dir, "legend.js"), "r") as f:
        legend = f.read()
    with open(os.path.join(src_dir, "modal.js"), "r") as f:
        modal = f.read()
    with open(os.path.join(src_dir, "ptN2.js"), "r") as f:
        pt_n2 = f.read()
    with open(os.path.join(src_dir, "search.js"), "r") as f:
        search = f.read()
    with open(os.path.join(src_dir, "svg.js"), "r") as f:
        svg = f.read()

    #grab the style
    with open(os.path.join(style_dir, "awesomplete.css"), "r") as f:
        awesomplete_style = f.read()
    with open(os.path.join(style_dir, "partition_tree.css"), "r") as f:
        partition_tree_style = f.read()
    with open(os.path.join(style_dir, "fontello.woff"), "rb") as f:
        encoded_font = str(base64.b64encode(f.read()).decode("ascii"))

    #grab the index.html
    with open(os.path.join(vis_dir, "index.html"), "r") as f:
        index = f.read()

    #grab the model viewer data
    model_viewer_data = 'var modelData = %s' % json.dumps(_get_viewer_data(problem_or_filename, include_wards))

    #add the necessary HTML tags if we aren't embedding
    if not embeddable:
        index = html_begin_tags + index + html_end_tags

    #put all style and JS into index
    index = index.replace('{{awesomplete_style}}', awesomplete_style)
    index = index.replace('{{partition_tree_style}}', partition_tree_style)
    index = index.replace('{{fontello}}', encoded_font)
    index = index.replace('{{d3_lib}}', d3)
    index = index.replace('{{awesomplete_lib}}', awesomplete)
    index = index.replace('{{vk_beautify_lib}}', vk_beautify)
    index = index.replace('{{model_data}}', model_viewer_data)
    index = index.replace('{{constants_lib}}', constants)
    index = index.replace('{{modal_lib}}', modal)
    index = index.replace('{{svg_lib}}', svg)
    index = index.replace('{{search_lib}}', search)
    index = index.replace('{{legend_lib}}', legend)
    index = index.replace('{{draw_lib}}', draw)
    index = index.replace('{{ptn2_lib}}', pt_n2)
    if draw_potential_connections:
        index = index.replace('{{draw_potential_connections}}', 'true')
    else:
        index = index.replace('{{draw_potential_connections}}', 'false')

    with open(outfile, 'w') as f:
        f.write(index)

    #open it up in the browser
    if show_browser:
        from cosapp.tools.problem_viewer.webview import webview
        webview(outfile)



"""
Utility classes for executing CoSApp simulation following file events.

Those classes use third-party `watchdog` package to observe file events.
"""
import logging
from threading import Timer
from typing import AnyStr, List, Union

from cosapp.systems import System

logger = logging.getLogger(__name__)


try:  # watchdog will be a conditional dependency
    from watchdog.events import (
        PatternMatchingEventHandler,
        FileCreatedEvent,
        FileModifiedEvent,
    )
    from watchdog.observers import Observer

    class WatchdogHandler(PatternMatchingEventHandler):
        """Define a watchdog class that allows to trigger actions on file events (creation,
        modification, etc.).
            
        Parameters
        ----------
        owner: System
            Owner of the watchdog. Actions will be triggered on him on events  
        folder: str
            Folder to supervise with the watchdog
        timeout: float, optional
            Define the max inactivity duration of the watchdog
        patterns: Union[List, AnyStr], optional
            Define the file patterns to monitor in the chosen path
        """

        def __init__(
            self,
            owner: System,
            folder: str,
            timeout: float = 30.0,
            patterns: Union[List, AnyStr] = "*.*",
        ):
            """`WatchdogHandler` class constructor
            
            Parameters
            ----------
            owner: System
                Owner of the watchdog. Actions will be triggered on him on events  
            folder: str
                Folder to supervise with the watchdog
            timeout: float, optional
                Define the max inactivity duration of the watchdog
            patterns: Union[List, AnyStr], optional
                Define the file patterns to monitor in the chosen path
            """
            super(WatchdogHandler, self).__init__(patterns=patterns)
            self._owner = owner
            self.timeout = timeout
            self.exit = False
            self.folder = folder

            self.__observer: Observer = None
            self.__timer: Timer = None
            self.__is_alive = False

        def compute(self, file_full_path: str) -> None:
            """Actions to do while an event is detected
            
            Parameters
            ----------
            file_full_path: str
                Gives access to the file that triggered an event
            """
            self._owner.run_drivers()

        def is_alive(self) -> bool:
            """Is the watcher alive?"""
            return self.__is_alive

        def reset(self) -> None:
            """Reset the watchdog by stopping the timer countdown and starting a new one"""
            if self.__observer is None:
                raise RuntimeError("Watcher has never been started.")
            self.__timer.cancel()
            self.__timer = Timer(self.timeout, self.stop)
            self.__timer.start()

        def stop(self) -> None:
            """Exits the watchdog monitoring"""
            if not self.__is_alive:
                logger.debug("Watcher not alive.")
                return

            self.__is_alive = False
            self.__timer.cancel()
            self.__observer.stop()
            self.__observer.join()

            logger.info(f"..trigger on {self._owner.name!r} is timeout")

        def start(self, time_step: float = 1.0) -> None:
            """Starts the watchdog monitoring

            Parameters
            ----------
            time_step: float, optional
                Gives the time step at which the watchdog will check events in the supervised folder
            """
            self.__observer = Observer()
            self.__observer.schedule(self, path=self.folder)

            if time_step > self.timeout:
                time_step = self.timeout

            logger.debug(
                f"> trigger created on {self._owner.name!r} with {self.timeout} seconds timeout.."
            )
            self.__observer.start()

            self.__timer = Timer(self.timeout, self.stop)
            self.__timer.start()
            self.__is_alive = True

    class FileCreationHandler(WatchdogHandler):
        """Specific watchdog class to monitor file creation events
            
        Parameters
        ----------
        owner: System
            Owner of the watchdog. Actions will be triggered on him on events  
        folder: str
            Folder to supervise with the watchdog
        timeout: float, optional
            Define the max inactivity duration of the watchdog
        patterns: Union[List, AnyStr], optional
            Define the file patterns to monitor in the chosen path
        """

        def on_created(self, event: FileCreatedEvent) -> None:  # when file is created
            """Actions to complete when a creation event is triggered
            
            Parameters
            ----------
            event
                Event that triggered the watchdog
            """
            logger.info(f" #creation of file {event.src_path!r} detected")
            self.compute(event.src_path)
            self.reset()
            logger.debug(
                f"> trigger restart on {self._owner.name!r} for another {self.timeout} seconds.."
            )

    class FileModificationHandler(WatchdogHandler):
        """Specific watchdog class to monitor file modification events.
            
        Parameters
        ----------
        owner: System
            Owner of the watchdog. Actions will be triggered on him on events  
        folder: str
            Folder to supervise with the watchdog
        timeout: float, optional
            Define the max inactivity duration of the watchdog
        patterns: Union[List, AnyStr], optional
            Define the file patterns to monitor in the chosen path
        """

        def on_modified(
            self, event: FileModifiedEvent
        ) -> None:  # when file is modified
            """Actions to complete when a modification event is triggered

            Parameters
            ----------
            event
                Event that triggered the watchdog
            """
            logger.info(f" #modification of file {event.src_path!r} detected")
            self.compute(event.src_path)
            self.reset()
            logger.debug(
                f"> trigger restart on {self._owner.name!r} for another {self.timeout} seconds.."
            )


except ImportError:
    logger.warning(
        "'watchdog' package is not installed. Files modification and creation trackers are not available."
    )



import pytest
import numpy as np

from cosapp.drivers import EulerExplicit
from cosapp.drivers.time.scenario import Scenario, Interpolator
from cosapp.core.eval_str import AssignString


@pytest.fixture(scope='function')
def case(two_tank_case):
    return two_tank_case(EulerExplicit, dt=0.1, time_interval=[0, 1])


@pytest.fixture(scope='function')
def scenario(case):
    driver = case[1]
    return Scenario('scenario', driver)


@pytest.fixture(scope='function')
def varying_mu_case(case):
    """Two-tank test case with exponentially decaying viscosity"""
    system, driver = case

    mu_0, mu_inf, tau = 1e-2, 2e-3, 200  # viscosity parameters
    mu = lambda t: mu_inf + (mu_0 - mu_inf) * np.exp(-t / tau)

    scenario = Scenario.make('scenario',
        driver = driver,
        init = {
            'tank1.height': '10 * pipe.D',
            'tank2.height': 'tank1.height + 1',
        },
        values = {
            'pipe.D': 0.3,
            'tank1.area': '5',
            'pipe.L': '50 * pipe.D',
            'pipe.mu': f'{mu_inf} + ({mu_0} - {mu_inf}) * exp(-t / {tau})',
        },
    )
    return system, driver, scenario, mu


def test_Scenario__init__(case):
    system, driver = case
    scenario = Scenario('scenario', driver)
    assert scenario.owner is driver
    assert scenario.context is system
    assert scenario.init_values == []
    assert scenario.case_values == []
    with pytest.raises(TypeError):
        scenario.owner = 'foo'


@pytest.mark.parametrize("init, expected", [
    (
        {'tank1.height': 1.2, 'tank2.height': 0.5},
        dict(content=['tank1.height = 1.2', 'tank2.height = 0.5'], n_const=2)
    ),
    (
        {'tank2.height': 0.5, 'tank1.height': 1.2},
        dict(content=['tank1.height = 1.2', 'tank2.height = 0.5'], n_const=2)
    ),
    (
        {'tank2.height': 'tank1.height + 1', 'tank1.height': 2, },
        dict(content=['tank1.height = 2', 'tank2.height = tank1.height + 1'], n_const=1)
    ),
    (
        {'tank1.height': '10 * pipe.D', 'tank2.height': 0.5, },
        dict(content=['tank2.height = 0.5', 'tank1.height = 10 * pipe.D'], n_const=1)
    ),
    (
        {'tank1.height': 'exp(1.5)', 'tank2.height': '10 * pipe.D', 'pipe.D': 0.25},
        dict(content=['pipe.D = 0.25', 'tank1.height = exp(1.5)', 'tank2.height = 10 * pipe.D'], n_const=2)
    ),
    (
        {'tank1.height': 'pipe.L / 2', 'tank2.height': '10 * pipe.D', 'pipe.D': 0.25},
        dict(content=['pipe.D = 0.25', 'tank1.height = pipe.L / 2', 'tank2.height = 10 * pipe.D'], n_const=1)
    ),
    (
        {'tank1.height': 'pipe.L / 2', 'tank2.height': '10 * pipe.D'},
        dict(content=['tank1.height = pipe.L / 2', 'tank2.height = 10 * pipe.D'], n_const=0)
    ),
    # Erroneous cases:
    ({'foo.bar': 2}, dict(error=AttributeError, match=r"'foo\.bar' is not known in system")),
    ({'pipe.k': 2}, dict(error=ValueError, match="Only variables in input ports can be used as boundaries")),
])
def test_Scenario_set_init(scenario, init, expected):
    error = expected.get('error', None)

    if error is None:
        scenario.set_init(init)
        assert all(isinstance(assignment, AssignString) for assignment in scenario.init_values)

        n_const = expected['n_const']
        content = expected['content']
        const = content[:n_const]
        nonconst = content[n_const:]
        init_conds = [str(assignment) for assignment in scenario.init_values]
        # First elements, expected to be constants init_conds 
        assert len(init_conds[:n_const]) == n_const
        assert set(init_conds[:n_const]) == set(const)
        assert all(assignment.constant for assignment in scenario.init_values[:n_const])
        # End init_conds (unordered, from dict entry)
        assert len(init_conds[n_const:]) == len(nonconst)
        assert set(init_conds[n_const:]) == set(nonconst)

    else:
        pattern = expected.get('match', None)
        with pytest.raises(error, match=pattern):
            scenario.set_init(init)


@pytest.mark.parametrize("values, expected", [
    (
        {'pipe.D': 0.3, 'pipe.L': '5'},
        dict(init=['pipe.L = 5', 'pipe.D = 0.3'], values=[],)
    ),
    (
        {'pipe.D': 0.3, 'pipe.L': '50 * pipe.D'},
        dict(init=['pipe.D = 0.3'], values=['pipe.L = 50 * pipe.D'],)
    ),
    (
        {
            'tank1.area': 2,
            'pipe.D': 0.3,
            'pipe.L': '50 * pipe.D',
            'pipe.mu': '{vinf} + ({v0} - {vinf}) * exp(-t / {tau})'.format(v0=1e-2, vinf=2e-3, tau=200),
        },
        dict(
            init=['pipe.D = 0.3', 'tank1.area = 2'],
            values=[
                'pipe.L = 50 * pipe.D',
                'pipe.mu = {vinf} + ({v0} - {vinf}) * exp(-t / {tau})'.format(v0=1e-2, vinf=2e-3, tau=200)
            ],
        )
    ),
    # Erroneous cases:
    ({'foo.bar': 2}, dict(error=AttributeError, match=r"'foo\.bar' is not known in system")),
    ({'pipe.k': 2}, dict(error=ValueError, match="Only variables in input ports can be used as boundaries")),
])
def test_Scenario_set_values(scenario, values, expected):
    error = expected.get('error', None)

    if error is None:
        scenario.set_values(values)
        assert all(isinstance(assignment, AssignString) for assignment in scenario.case_values)
        assert all(not assignment.constant for assignment in scenario.case_values)

        init_conds = [str(assignment) for assignment in scenario.init_values]
        time_conds = [str(assignment) for assignment in scenario.case_values]

        # Test time-dependent boundary conditions
        assert set(time_conds) == set(expected['values'])
        assert set(init_conds) == set(expected['init'])

    else:
        pattern = expected.get('match', None)
        with pytest.raises(error, match=pattern):
            scenario.set_values(values)


def test_Scenario_add_init(case):
    system, driver = case

    scenario = Scenario.make('scenario',
        driver = driver,
        values = {},
        init = {
            'tank1.height': '10 * pipe.D',
        },
    )
    assert len(scenario.init_values) == 1
    assert str(scenario.init_values[0]) == 'tank1.height = 10 * pipe.D'

    scenario.add_init({'tank2.height': 'tank1.height + 1'})
    assert len(scenario.init_values) == 2
    assert str(scenario.init_values[1]) == 'tank2.height = tank1.height + 1'

    scenario.add_init({'tank2.area': '1.3'})
    assert len(scenario.init_values) == 3
    assert str(scenario.init_values[0]) == 'tank2.area = 1.3'
    assert str(scenario.init_values[1]) == 'tank1.height = 10 * pipe.D'
    assert str(scenario.init_values[2]) == 'tank2.height = tank1.height + 1'


def test_Scenario_add_values(case):
    system, driver = case
    mu_0, mu_inf, tau = 1e-2, 2e-3, 200

    scenario = Scenario.make('scenario',
        driver = driver,
        values = {
            'tank1.area': '5',  # constant
            'pipe.D': 0.3,      # constant
            'pipe.mu': f'{mu_inf} + ({mu_0} - {mu_inf}) * exp(-t / {tau})',
        },
        init = {
            'tank1.height': '10 * pipe.D',
            'tank2.height': 'tank1.height + 1',
        },
    )

    assert len(scenario.init_values) == 4
    assert len(scenario.case_values) == 1

    init_conds = [str(assignment) for assignment in scenario.init_values]
    assert set(init_conds[:2]) == {'tank1.area = 5', 'pipe.D = 0.3'}
    assert set(init_conds[2:]) == {
        'tank1.height = 10 * pipe.D',
        'tank2.height = tank1.height + 1'
    }

    case_conds = [str(assignment) for assignment in scenario.case_values]
    assert set(case_conds) == {
            f'pipe.mu = {mu_inf} + ({mu_0} - {mu_inf}) * exp(-t / {tau})',
    }

    # Add non-const boundary condition
    scenario.add_values({'tank2.area': '10 * pipe.D**2'})
    assert len(scenario.init_values) == 4
    assert len(scenario.case_values) == 2
    assert str(scenario.case_values[-1]) == 'tank2.area = 10 * pipe.D**2'

    # Add constant boundary condition
    scenario.add_values({'pipe.L': 3.4})
    assert len(scenario.init_values) == 5
    assert len(scenario.case_values) == 2
    assert str(scenario.init_values[0]) == 'pipe.L = 3.4'


def test_Scenario_add_values_tabulated(case):
    system, driver = case
    mu_0, mu_inf, tau = 1e-2, 2e-3, 200

    scenario = Scenario.make('scenario',
        driver = driver,
        values = {
            'tank1.area': '5',  # constant
            'pipe.D': 0.3,      # constant
            'pipe.mu': Interpolator(data=[(0, mu_0), (1, mu_0), (1.1, mu_inf), (1000, mu_inf)]),
        },
        init = {
            'tank1.height': '10 * pipe.D',
            'tank2.height': 'tank1.height + 1',
        },
    )

    assert len(scenario.init_values) == 4
    assert len(scenario.case_values) == 1

    init_conds = [str(assignment) for assignment in scenario.init_values]
    assert set(init_conds[:2]) == {'tank1.area = 5', 'pipe.D = 0.3'}
    assert set(init_conds[2:]) == {
        'tank1.height = 10 * pipe.D',
        'tank2.height = tank1.height + 1'
    }

    case_conds = [str(assignment) for assignment in scenario.case_values]
    assert set(case_conds) == {
            'pipe.mu = Interpolator(t)',
    }


@pytest.mark.parametrize("case_data, expected", [
    (
        dict(
            init = {'tank1.height': '10 * pipe.D', 'tank2.height': 'tank1.height + 1'},
            values = {'pipe.D': 0.3, 'tank1.area': '5'},
        ),
        dict(
            init = dict(
                content=[
                    'pipe.D = 0.3', 'tank1.area = 5',  # const assignments passed as init conditions
                    'tank1.height = 10 * pipe.D', 'tank2.height = tank1.height + 1'
                ],
                n_const=2
            ),
            values = [],
        )
    ),
    (
        dict(
            init = {'tank1.height': '10 * pipe.D', 'tank2.height': 'tank1.height + 1'},
            values = {
                'pipe.D': 0.3,
                'tank1.area': '5',
                'pipe.L': '50 * pipe.D',
                'pipe.mu': '{vinf} + ({v0} - {vinf}) * exp(-t / {tau})'.format(v0=1e-2, vinf=2e-3, tau=200),
            },
        ),
        dict(
            init = dict(
                content=[
                    'pipe.D = 0.3', 'tank1.area = 5',  # const assignments passed as init conditions
                    'tank1.height = 10 * pipe.D', 'tank2.height = tank1.height + 1'
                ],
                n_const=2
            ),
            values = [
                'pipe.L = 50 * pipe.D',
                'pipe.mu = {vinf} + ({v0} - {vinf}) * exp(-t / {tau})'.format(v0=1e-2, vinf=2e-3, tau=200),
            ],
        )
    ),
])
def test_Scenario_factory(case, case_data, expected):
    system, driver = case
    scenario = Scenario.make('scenario',
        driver = driver,
        init = case_data['init'],
        values = case_data['values'],
    )

    assert scenario.owner is driver
    assert scenario.context is system
    assert isinstance(scenario, Scenario)
    assert all(isinstance(assignment, AssignString) for assignment in scenario.init_values)
    assert all(isinstance(assignment, AssignString) for assignment in scenario.case_values)
    assert all(assignment.eval_context is system for assignment in scenario.init_values)
    assert all(assignment.eval_context is system for assignment in scenario.case_values)

    init_conds = [str(assignment) for assignment in scenario.init_values]
    time_conds = [str(assignment) for assignment in scenario.case_values]

    # Test time-dependent boundary conditions
    assert set(time_conds) == set(expected['values'])

    # Test initial conditions
    content = expected['init']['content']
    n_const = expected['init']['n_const']
    const = content[:n_const]
    nonconst = content[n_const:]
    # First elements, expected to be constants init_conds 
    assert len(init_conds[:n_const]) == n_const
    assert set(init_conds[:n_const]) == set(const)
    assert all(assignment.constant for assignment in scenario.init_values[:n_const])
    # End init_conds (unordered, from dict entry)
    assert len(init_conds[n_const:]) == len(nonconst)
    assert set(init_conds[n_const:]) == set(nonconst)


def test_Scenario_apply_values(varying_mu_case, set_master_system):
    """Test methods `apply_init_values()` and `update_values()`"""
    system, driver, scenario, mu = varying_mu_case

    system.tank1.height = system.tank2.height = 1
    system.tank1.area = system.tank2.area = 1
    system.pipe.D = 1
    system.pipe.L = 1
    system.pipe.mu = 0.1

    scenario.apply_init_values()
    assert system.pipe.D == 0.3
    assert system.tank1.height == 3
    assert system.tank2.height == 4
    assert system.tank1.area == 5
    assert system.tank2.area == 1
    # Check that non-const boundary conditions have not been applied yet
    assert system.pipe.L == 1
    assert system.pipe.mu == 0.1

    # Apply time-dependent boundary conditions
    driver._set_time(0.)
    assert system.time == 0
    scenario.update_values()
    assert system.pipe.L == 15
    assert system.pipe.mu == mu(system.time)

    driver._set_time(10.)
    assert system.time == 10
    scenario.update_values()
    assert system.pipe.L == 15
    assert system.pipe.mu == mu(system.time)


def test_Scenario_clear_init(varying_mu_case):
    scenario = varying_mu_case[2]
    assert len(scenario.init_values) == 4
    assert len(scenario.case_values) == 2
    
    scenario.clear_init()
    assert len(scenario.init_values) == 0
    assert len(scenario.case_values) == 2


def test_Scenario_clear_values(varying_mu_case):
    scenario = varying_mu_case[2]
    assert len(scenario.init_values) == 4
    assert len(scenario.case_values) == 2
    
    scenario.clear_values()
    assert len(scenario.init_values) == 4
    assert len(scenario.case_values) == 0



import sys
from pathlib import Path

import pytest


@pytest.fixture
def test_library():
    library_path = Path(__file__).parent / "library" / "systems"

    # Add path to allow System to find the component
    sys.path.append(str(library_path))
    try:
        yield library_path
    finally:
        # Undo path modification
        sys.path.remove(str(library_path))

@pytest.fixture
def test_data():
    return Path(__file__).parent / "data"



"""Test time observer pattern used in CoSApp"""
import pytest

from cosapp.core.time import TimeManager, UniversalClock, TimeObserver


@pytest.fixture(scope="function", autouse=True)
def final_check():
    yield None
    # teardown
    assert UniversalClock().n_observers == 0


@pytest.fixture(scope="function")
def clock():
    return UniversalClock()


class MockTimeObserver(TimeObserver):
    """Concrete mock-up of TimeObserver"""
    def _update(self, dt):
        pass


@pytest.mark.parametrize("args", [[], [0], [0.5], [-10]])
def test_TimeManager__init__(args):
    manager = TimeManager(*args)
    assert manager.time == (args[0] if args else 0)
    assert manager.n_observers == 0


def test_UniversalClock__init__():
    clock = UniversalClock()
    assert clock.time == 0
    assert clock.n_observers == 0
    clock.time = 0.2
    assert clock.time == 0.2
    clock.reset()
    assert clock.time == 0


def test_UniversalClock_unicity():
    clock1 = UniversalClock()
    clock2 = UniversalClock()
    assert clock1 is clock2


def test_TimeObserver_explicit_delete(clock):
    observer = MockTimeObserver(True)
    assert clock.n_observers == 1
    del observer
    assert clock.n_observers == 0


@pytest.mark.parametrize("sign_in", [True, False])
def test_TimeObserver__init__(sign_in, clock):
    observer = MockTimeObserver(sign_in)
    assert hasattr(observer, 'time')
    assert observer.time == clock.time
    assert observer.observes() == sign_in
    assert clock.n_observers == (1 if sign_in else 0)


def test_TimeObserver__delete__(clock):
    observers = [MockTimeObserver(True) for _ in range(10)]
    assert clock.n_observers == len(observers)

    for _ in range(5):
        MockTimeObserver(True)  # create unreferenced, temporary objects
    assert clock.n_observers == len(observers)

    for _ in range(5):
        tmp = MockTimeObserver(True)  # referenced temporary objects
        assert clock.n_observers == len(observers) + 1
    assert clock.n_observers == len(observers) + 1
    del tmp
    assert clock.n_observers == len(observers)




from typing import Optional

from cosapp.core.numerics.distributions.uniform import Uniform
from cosapp.systems import SystemFamily, System
from cosapp.tests.library.ports import XPort, X2Port, X3Port
import numpy as np


class MultiplyFamily(SystemFamily):

    def __init__(self, name: str):
        super(MultiplyFamily, self).__init__(name)

        self.family_name = 'Multiply'

        self.modelings.add('Multiply1', fidelity=0.99, cost=0.1)
        self.modelings.add('Multiply2', fidelity=1, cost=0.3)
        self.modelings.add('Multiply3', fidelity=1.01, cost=0.9)


class Multiply1(MultiplyFamily):

    def setup(self):
        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_inward('K1', 5., distribution=Uniform(worst=-1, best=+1))
        self.add_output(XPort, 'p_out', {'x': 1.})

    def compute(self):
        self.p_out.x = self.p_in.x * self.K1

    def Multiply1_to_Multiply2(self):
        rtn = Multiply2(self.name)
        rtn.parent = self.parent

        rtn.K1 = self.K1 ** 0.5
        rtn.K2 = rtn.K1

        self.update_connections(rtn)
        return rtn

    def Multiply1_to_Multiply3(self):
        rtn = Multiply3(self.name)
        rtn.parent = self.parent

        rtn.K1 = self.K1 ** (1. / 3.)
        rtn.K2 = rtn.K1
        rtn.K3 = rtn.K1

        self.update_connections(rtn)
        return rtn


class Multiply2(MultiplyFamily):

    def setup(self):
        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_inward('K1', 5., distribution=Uniform(worst=-0.1, best=+0.1))
        self.add_inward('K2', 5., distribution=Uniform(worst=-0.2, best=+0.2))
        self.add_output(XPort, 'p_out', {'x': 1.})

        self.add_outward('Ksum', 0.)

    def compute(self):
        self.p_out.x = self.p_in.x * self.K1 * self.K2
        self.Ksum = self.K1 + self.K2

    def Multiply2_to_Multiply1(self):
        rtn = Multiply1(self.name)
        rtn.parent = self.parent

        rtn.K1 = self.K1 * self.K2

        self.update_connections(rtn)
        return rtn

    def Multiply2_to_Multiply3(self):
        rtn = Multiply3(self.name)
        rtn.parent = self.parent

        rtn.K1 = (self.K1 * self.K2) ** (1 / 3)
        rtn.K2 = rtn.K1
        rtn.K3 = rtn.K1

        self.update_connections(rtn)
        return rtn


class Multiply2Derivative(MultiplyFamily):

    def setup(self):
        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_inward('K1', 5.)
        self.add_inward('K2', 5.)
        self.add_output(XPort, 'p_out', {'x': 1.})

        self.add_inward('dK1_dK2', 0.)
        self.add_derivative('inwards.dK1_dK2', 'inwards.K1', 'inwards.K2')

    def compute(self):
        self.p_out.x = self.p_in.x * self.K1 * self.K2


class Multiply3(MultiplyFamily):

    def setup(self):
        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_inward({'K1': 5., 'K2': 5., 'K3': 5.})
        self.add_output(XPort, 'p_out', {'x': 1.})

    def compute(self):
        self.p_out.x = self.p_in.x * self.K1 * self.K2 * self.K3

    def Multiply3_to_Multiply1(self):
        rtn = Multiply1(self.name)
        rtn.parent = self.parent

        rtn.K1 = self.K1 * self.K2 * self.K3

        self.update_connections(rtn)
        return rtn

    def Multiply3_to_Multiply2(self):
        rtn = Multiply2(self.name)
        rtn.parent = self.parent

        rtn.K1 = (self.K1 * self.K2 * self.K3) ** 0.5
        rtn.K2 = rtn.K1

        self.update_connections(rtn)
        return rtn


class Multiply4(System):

    def setup(self):
        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_inward('K1', 2., valid_range=(1., 6.))
        self.add_inward('K2', 1., distribution=Uniform(worst=-0.1, best=0.1))
        self.add_output(XPort, 'p_out', {'x': 1.})

    def compute(self):
        self.p_out.x = self.p_in.x * self.K1 * self.K2


class MultiplySystem(System):
    def setup(self):
        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_output(XPort, 'p_out', {'x': 1.})

        self.add_child(Multiply1('mult1'))
        self.connect(self.p_in, self.mult1.p_in)
        self.connect(self.p_out, self.mult1.p_out)

        self.exec_order = ['mult1']


class MultiplySystem2(System):
    def setup(self):
        self.add_child(Multiply1('mult1'))
        self.add_child(Multiply1('mult2'))

        self.connect(self.mult1.p_out, self.mult2.p_in)

        self.exec_order = ['mult1', 'mult2']


class MultiplyVector2(System):
    def setup(self):
        self.add_input(X2Port, 'p_in', {'x1': 1., 'x2': 1.})
        self.add_inward('k1', 5.)
        self.add_inward('k2', 5.)
        self.add_output(XPort, 'p_out', {'x': 1.})

    def compute(self):
        self.p_out.x = self.k1 * self.p_in.x1 + self.k2 * self.p_in.x2


class MultiplyVector3(System):

    def setup(self):
        self.add_input(X3Port, 'p_in', {'x1': 1., 'x2': 1., 'x3': 1.})
        self.add_inward('k1', 5.)
        self.add_inward('k2', 5.)
        self.add_inward('k3', 5.)
        self.add_output(XPort, 'p_out', {'x': 1.})

    def compute(self):
        self.p_out.x = self.k1 * self.p_in.x1 + self.k2 * self.p_in.x2 + self.k3 * self.p_in.x3


class ExponentialLoad(System):
    """ExponentialLoad expression

    $load = \\exp(a * (p_in.x - b)) + c$

    Default values (a, b, c) = (1, 0, 0) => load = exp(p_in.x)
    """

    def setup(self, **kwargs):
        self.add_input(XPort, 'p_in')
        self.add_inward('a', 1)
        self.add_inward('b', 0)
        self.add_inward('c', 0)
        self.add_outward('load_', 0.)
        self.add_output(XPort, 'p_out')

    def compute(self):
        self.load_ = np.exp(self.a * (self.p_in.x - self.b)) + self.c
        self.p_out.x = self.p_in.x


class NonLinear1(System):

    def setup(self):
        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_inward('k1', 5.)
        self.add_inward('k2', 5.)
        self.add_output(XPort, 'p_out', {'x': 1.})

    def compute(self):
        self.p_out.x = self.k1 * self.p_in.x ** self.k2


class NonLinear3(System):

    def setup(self):
        self.add_input(X3Port, 'p_in', {'x1': 1., 'x2': 1., 'x3': 1.})
        self.add_inward('k1', 5.)
        self.add_inward('k2', 5.)
        self.add_inward('k3', 5.)
        self.add_output(XPort, 'p_out', {'x': 1.})

    def compute(self):
        self.p_out.x = self.k1 * self.p_in.x1 ** 2 + self.k2 / self.p_in.x2 ** 0.5 + self.k3 * self.p_in.x3


class Merger(System):

    def setup(self):
        self.add_input(XPort, 'p1_in', {'x': 1.})
        self.add_input(XPort, 'p2_in', {'x': 1.})
        self.add_output(XPort, 'p_out', {'x': 1.})

    def compute(self):
        self.p_out.x = self.p1_in.x + self.p2_in.x


class Splitter(System):

    def setup(self):
        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_inward('split_ratio', 0.1)
        self.add_output(XPort, 'p1_out', {'x': 1.})
        self.add_output(XPort, 'p2_out', {'x': 1.})

    def compute(self):
        self.p1_out.x = self.p_in.x * self.split_ratio
        self.p2_out.x = self.p_in.x * (1 - self.split_ratio)


class IterativeNonLinear(System):

    def setup(self):
        self.add_child(NonLinear1('nonlinear'))
        self.add_child(Multiply2('mult2'))
        self.add_child(Merger('merger'))
        self.add_child(Splitter('splitter'))

        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_output(XPort, 'p_out', {'x': 1.})

        self.connect(self.p_in, self.merger.p1_in)
        self.connect(self.nonlinear.p_out, self.merger.p2_in)
        self.connect(self.merger.p_out, self.mult2.p_in)
        self.connect(self.mult2.p_out, self.splitter.p_in)
        self.connect(self.splitter.p1_out, self.nonlinear.p_in)
        self.connect(self.splitter.p2_out, self.p_out)

        self.exec_order = ['merger', 'mult2', 'splitter', 'nonlinear']


class IterativeNonLinearDerivative(System):

    def setup(self):
        self.add_child(NonLinear1('nonlinear'))
        self.add_child(Multiply2Derivative('mult2'))
        self.add_child(Merger('merger'))
        self.add_child(Splitter('splitter'))

        self.add_input(XPort, 'p_in', {'x': 1.})
        self.add_output(XPort, 'p_out', {'x': 1.})

        self.connect(self.p_in, self.merger.p1_in)
        self.connect(self.nonlinear.p_out, self.merger.p2_in)
        self.connect(self.merger.p_out, self.mult2.p_in)
        self.connect(self.mult2.p_out, self.splitter.p_in)
        self.connect(self.splitter.p1_out, self.nonlinear.p_in)
        self.connect(self.splitter.p2_out, self.p_out)

        self.exec_order = ['merger', 'mult2', 'splitter', 'nonlinear']



"""
Module defining the Slot class.
"""
import types
from typing import Any, Callable, Optional
import weakref


class Slot:
    """
    A slot is a callable object that manages a connection to a signal.
    If weak is true or the slot is a subclass of weakref.ref, the slot
    is automatically de-referenced to the called function.
    """

    __slots__ = ["__weak", "__slot"]

    def __init__(self, slot: Callable, weak: bool=True):
        if not callable(slot):
            raise TypeError("A `Slot` instance must be defined from a callable object.")
            
        self.__weak = bool(weak) or isinstance(slot, weakref.ref)
        if weak and not isinstance(slot, weakref.ref):
            if isinstance(slot, types.MethodType):
                slot = weakref.WeakMethod(slot)
            else:
                slot = weakref.ref(slot)
        self.__slot = slot

    @property
    def is_alive(self) -> bool:
        """
        Return `True` if slot is alive, `False` otherwise.
        """
        return (not self.__weak) or (self.__slot() is not None)

    @property
    def func(self) -> Callable:
        """
        Return the function called by slot.
        """
        return self.__slot() if self.__weak else self.__slot

    def __call__(self, **kwargs) -> Optional[Any]:
        """
        Execute slot.
        """
        func = self.func
        if func is not None:
            return func(**kwargs)

    def __eq__(self, other) -> bool:
        """
        Compare this slot to another.
        """
        if isinstance(other, Slot):
            return self.func == other.func
        else:
            return self.func == other

    def __repr__(self) -> str:
        f = self.func
        s = "dead" if f is None else repr(f)
        return f"<cosapp.core.signal.Slot: {s}>"



import os
import unittest
from shutil import copy2, rmtree
from tempfile import mkdtemp

import pytest

nbconvert = pytest.importorskip("nbconvert")

from nbconvert.preprocessors.execute import executenb
from nbformat import read, current_nbformat

test_folder = os.path.dirname(__file__)
tutorials_folder = os.path.join(test_folder, '../../docs/tutorials')
if not os.path.exists(tutorials_folder):  # For distributed cosapp package
    tutorials_folder = os.path.join(test_folder, '../tutorials')


@pytest.fixture(scope="module")
def copy_tutorials():
    tutorials_tmp = mkdtemp()

    if os.path.exists(tutorials_folder):
        curdir = tutorials_folder
        for f in os.listdir(tutorials_folder):
            curfile = os.path.join(curdir, f)
            if os.path.isfile(curfile) and os.path.splitext(curfile)[1] in (".ipynb", ".py"):
                copy2(curfile, tutorials_tmp)

    yield tutorials_tmp
    
    # teardown
    rmtree(tutorials_tmp)


@pytest.mark.notebook
@pytest.mark.parametrize("filename",
    [file for file in os.listdir(tutorials_folder) if os.path.splitext(file)[1] == '.ipynb'])
def test_all(copy_tutorials, filename):
    fullpath = os.path.join(copy_tutorials, filename)
    assert os.path.exists(fullpath), f"Tutorial {filename} was not copied."

    with open(fullpath) as fnb:
        nb = read(fnb, current_nbformat)

    # The execution of the notebook should not raise
    # nbconvert.preprocessors.execute.CellExecutionError
    executenb(nb, cwd=copy_tutorials, timeout=60)



"""
System tuned to support meta-models.
"""
import json
from pathlib import Path
from typing import Any, Dict, Optional, Union, Tuple, Type

import numpy
import pandas

from cosapp.ports.port import ExtensiblePort
from cosapp.ports.enum import Scope
from cosapp.systems.system import System
from cosapp.systems.surrogate_models.nearest_neighbor import NearestNeighbor
from cosapp.systems.surrogate_models.surrogate_model import SurrogateModel
# TODO allow keywords arguments to be passed to the surrogate constructor
from cosapp.utils.helpers import is_numerical, is_number


class MetaSystem(System):
    """Based :py:class:`~cosapp.systems.system.System` to insert meta-system inside CoSApp.

    Parameters
    ----------
    data : str or pandas.DataFrame
    mapping : Dict[str, str], optional
    default_model : Type[SurrogateModel], optional
        Default surrogate model; default NearestNeighbor with RBF interpolator
    name : str, optional
        Name of the `System`; default 'undefined'

    Attributes
    ----------
    models : dict[str, SurrogateModel]
        Dictionary of surrogate model for each output variables
    _trained : bool
        Are the surrogate models trained?
    _default_model : Type[SurrogateModel]
        Default surrogate model to use
    _input_names : list of str
        List of port variable names (`port`.`variable`)

    Examples
    --------

    As for classical :py:class:`~cosapp.systems.system.System`, this class needs to be subclassed to be meaning full.

    >>> class XPort(Port):
    >>>     def setup(self):
    >>>         self.add_variable('x')
    >>>
    >>> class OutPort(Port):
    >>>     def setup(self):
    >>>         self.add_variable('sin')
    >>>         self.add_variable('cos')
    >>>
    >>> class MetaA(MetaSystem):
    >>>     def setup(self):
    >>>         self.add_input(XPort, 'in_')
    >>>         self.add_output(OutPort, 'out', model=NearestNeighbor)

    """

    __slots__ = ('_default_model', '_input_names', '_trained', 'models')

    def __init__(self,
        name: str,
        data: Union[str, pandas.DataFrame],
        mapping: Optional[Dict[str, str]] = None,
        default_model: Type[SurrogateModel] = NearestNeighbor,
        **kwargs):
        """Class constructor"""
        self._trained = False
        self._default_model = default_model
        self.models = dict()
        self._input_names = None

        kwargs['data'] = data
        kwargs['mapping'] = mapping
        super().__init__(name, **kwargs)  # here _initialize will be called then user setup

        self._input_names = self._get_models_inputs()

    def _get_models_inputs(self) -> list:
        # TODO Fred this does not smell good, sequence only allow in System.INWARDS port...
        inputs = list()
        sequences = list()

        for port in self.inputs.values():
            if port.name == MetaSystem.INWARDS:
                ignored = ('training_data', )

                for varname in filter(lambda n: n not in ignored, port):
                    value = port[varname]
                    if not is_numerical(value):
                        continue
                    elif is_number(value):
                        inputs.append(f"{port.name}.{varname}")
                    else:
                        sequences.append(varname)
            else:
                inputs.extend([f"{port.name}.{varname}" for varname in port if is_number(port[varname])])

        self._locked = False
        # TODO Fred We should not change the System interface but work with group of model instead for sequences
        for s in sequences:
            for i, val in enumerate(self.inputs[MetaSystem.INWARDS][s]):
                name = f"{s}_index{i}"
                self.add_inward(name, val)
                inputs.append(f"{MetaSystem.INWARDS}.{name}")
            self.split_sequences_in_training_data(MetaSystem.INWARDS, s)
        self._locked = True

        return inputs

    def _initialize(self,
        data: Union[str, pandas.DataFrame] = None,
        mapping: Dict[str, str] = None,
        **kwargs) -> Dict[str, Any]:
        """Hook method to add `Module` member before calling `setup` method.

        Parameters
        ----------
        data : str or pandas.DataFrame
            The training data or a filepath containing them
        mapping : Dict[str, str]
            Name mapping between training data columns and port variables.
        kwargs : Dict[str, Any]
            Other keywords arguments to pass to `setup` method

        Returns
        -------
        Dict[str, Any]
            ``System.__init__`` optional keywords arguments not consumed by this method.
        """

        def read_data(data):
            data_path = Path(data)
            if not data_path.is_file():
                raise FileNotFoundError(f"File {data_path.name} does not exist.")

            def read_json(path: Path) -> pandas.DataFrame:
                with path.open() as file:
                    tmp_dict = json.load(file)
                if set(tmp_dict.keys()) == {'schema', 'data'}:
                    # File is in table format
                    data, schema = tmp_dict['data'], tmp_dict['schema']
                    return pandas.DataFrame(data).set_index(schema['primaryKey'])
                else:
                    return pandas.DataFrame(tmp_dict)

            readers = {
                '.csv': pandas.read_csv,
                '.xlsx': pandas.read_excel,
                '.json': read_json,
            }

            def read_unknown_sep(path: Path) -> pandas.DataFrame:
                return pandas.read_csv(path, sep=None, engine='python')

            data = readers.get(data_path.suffix, read_unknown_sep)(data_path)
            return data

        if isinstance(data, str):
            data = read_data(data)
        elif not isinstance(data, pandas.DataFrame):
            raise TypeError(
                f"data must be a str or a pandas.DataFrame: got {type(data).__name__}."
            )

        if mapping is not None:
            data = data.rename(columns=mapping)  # Rename columns

        self.add_inward('training_data', data, desc="Training data")
        return kwargs

    def _train(self) -> None:
        """Train the model using the provided training data."""

        for model_name, model in self.models.items():
            if model_name not in self.training_data.columns:
                raise AttributeError(
                    f"Name {model_name!r} not found in training data of MetaModel {self.name!r}"
                )
            model.train(
                self.training_data.loc[:, self._input_names].values,
                self.training_data.loc[:, model_name].values.reshape((-1, 1))
            )
        self._trained = True

    def add_output(self,
        port_class: "Type[Port]",
        name: str,
        variables: Optional[Dict[str, Any]] = None,
        model: Optional[Union[Type[SurrogateModel], Dict[str, SurrogateModel]]] = None,
    ) -> ExtensiblePort:
        """Add an output `Port` to the `System`.

        This function cannot be called outside `System.setup`.

        Parameters
        ----------
        port_class: type
            Class of the `Port` to create
        name : str
            `Port` name
        variables : Dict[str, Any], optional
            Dictionary of initial values (default: None)
        model : Type[Surrogate] or Dict[str, SurrogateModel], optional
            Surrogate default model type or mapping of surrogate model with output names; default
            `NearestNeighbor`.

        Returns
        -------
        ExtensiblePort
            The created port

        Examples
        --------

        >>> class MyPort(Port):
        >>>     def setup(self):
        >>>         self.add_variable('x')
        >>>         self.add_variable('y')
        >>>
        >>> class MySystem(System):
        >>>     def setup(self):
        >>>         self.add_output(MyPort, 'output_x')
        >>>         self.add_output(MyPort, 'output_y', model=NearestNeighbor)
        >>>         self.add_output(MyPort, 'output_z', model={'x': NearestNeighbor,
        >>>                                                    'y': FloatKrigingSurrogate})
        """
        if not (isinstance(model, (dict, type(None))) or issubclass(model, SurrogateModel)):
            raise TypeError(
                f"Model should be a subclass of SurrogateModel or a dictionary; got {type(model)}."
            )
        port = super().add_output(port_class, name, variables)

        if model is None:
            model = {'default': self._default_model}
        elif isinstance(model, type):
            if issubclass(model, SurrogateModel):
                model = {'default': model}

        for variable in port:
            model_name = f"{name}.{variable}"
            surrogate_model = model.get(variable, model['default']())
            self.models[model_name] = surrogate_model

        return port

    def add_outward(self,
        definition: Union[str, Dict[str, Any]],
        value: Any = 1,
        unit: str = '',
        dtype: Optional[type] = None,
        valid_range: Optional[Tuple[Any, Any]] = None,
        invalid_comment: str = '',
        limits: Optional[Tuple[Any, Any]] = None,
        out_of_limits_comment: str = '',
        desc: str = "",
        scope: Scope = Scope.PUBLIC,
        model: Optional[Union[Type[SurrogateModel], Dict[str, SurrogateModel]]] = None,
    ) -> None:

        if not (isinstance(model, (dict, type(None))) or issubclass(model, SurrogateModel)):
            raise TypeError(
                f"Model should be a subclass of SurrogateModel or a dictionary; got {type(model)}."
            )

        super().add_outward(
            definition,
            value=value,
            unit=unit,
            dtype=dtype,
            valid_range=valid_range,
            invalid_comment=invalid_comment,
            limits=limits,
            out_of_limits_comment=out_of_limits_comment,
            desc=desc,
            scope=scope,
        )

        if model is None:
            model = {'default': self._default_model}
        elif isinstance(model, type):
            if issubclass(model, SurrogateModel):
                model = {'default': model}

        def add_model(variable: str):
            model_name = f"{MetaSystem.OUTWARDS}.{variable}"
            surrogate_model = model.get(variable, model['default']())
            self.models[model_name] = surrogate_model

        if isinstance(definition, str):
            var_name = definition
            if is_numerical(self.outputs[MetaSystem.OUTWARDS][var_name]):
                if is_number(self.outputs[MetaSystem.OUTWARDS][var_name]):
                    add_model(var_name)
                elif is_numerical(self.outputs[MetaSystem.OUTWARDS][var_name]):
                    for i, varname in enumerate(self.outputs[MetaSystem.OUTWARDS][var_name]):
                        self.add_outward(
                            f"{var_name}_index{i}",
                            value=varname,
                            desc=desc,
                            scope=scope,
                        )
                    self.split_sequences_in_training_data(MetaSystem.OUTWARDS, var_name)
        elif isinstance(definition, dict):
            for varname in definition:
                if is_number(self.outputs[MetaSystem.OUTWARDS][varname]):
                    add_model(varname)

    def split_sequences_in_training_data(self, port_name, var_name):
        name = f"{port_name}.{var_name}"
        training_data = self.inputs[MetaSystem.INWARDS]['training_data']
        if name in training_data.columns:
            temp = pandas.DataFrame(
                training_data[name].tolist(),
                index = training_data.index,
                columns = [f"{name}_index{i}" for i, _ in enumerate(self[name])]
            )
            self.inputs[MetaSystem.INWARDS]['training_data'] = pandas.concat([training_data, temp], axis=1)
        else:
            raise AttributeError(f"Name {name!r} not found in training data of MetaModel {self.name!r}")

    def compute(self) -> None:
        """Contains the customized `System` calculation."""
        if not self._trained:
            self._train()

        self.split_in_sequences()
        # build input vector
        input_vector = numpy.array([self[name] for name in self._input_names])

        for model_name, model in self.models.items():
            self[model_name] = float(model.predict(input_vector))

        self.rebuild_out_sequences()

    def split_in_sequences(self):
        for variable in self.inputs[MetaSystem.INWARDS]:
            var_name = f"{MetaSystem.INWARDS}.{variable}"
            if is_numerical(self[var_name]) and not is_number(self[var_name]):
                for i, val in enumerate(self[var_name]):
                    self[f"{var_name}_index{i}"] = val

    def rebuild_out_sequences(self):
        for name, port in self.outputs.items():
            for variable in port:
                var_name = f"{name}.{variable}"
                if is_numerical(self[var_name]) and not is_number(self[var_name]):
                    for i in range(len(self[var_name])):
                        self[var_name][i] = self[f"{var_name}_index{i}"]

    def update(self):
        self._trained = False  # Force model training



import pytest

import numpy as np
from numbers import Number
from cosapp.ports import Port
from cosapp.systems import System
from cosapp.core.time import UniversalClock


@pytest.fixture(scope='function')
def clock():
    yield UniversalClock()
    # teardown
    UniversalClock().reset()


@pytest.fixture
def set_master_system():
    """Ensure the System class variable master is properly restored"""
    System._System__master_set = True
    try:
        yield
    finally:
        System._System__master_set = False


def case_factory(system_cls, name, **kwargs):
    """Case factory used in test fixtures below"""
    def _test_objects(driver_cls, **options):
        system = system_cls(name, **kwargs)
        driver = system.add_driver(driver_cls(**options))
        return system, driver
    return _test_objects

# <codecell>

class ExpRampOde(System):
    """
    System representing function f(t) = a * (1 - exp(-t / tau)),
    through ODE: tau * f' + f = a
    """
    def setup(self):
        self.add_inward('a', 1.0)
        self.add_inward('tau', 1.0)

        self.add_outward('df_dt', 0.0)
        self.add_transient('f', der='df_dt', max_time_step='tau / 5')

    def compute(self):
        self.df_dt = (self.a - self.f) / self.tau

    def __call__(self, t) -> float:
        """Analytical solution at time t"""
        return self.a * (1 - np.exp(-t / self.tau))


@pytest.fixture(scope='function')
def ode_case_1():
    return case_factory(ExpRampOde, 'ode')

# <codecell>

class GaussianOde(System):
    """
    System representing function f(t) = a * exp(-0.5 * ((t - t0) / tau)^2)
    """
    def setup(self):
        self.add_inward('a', 1.0)
        self.add_inward('t0', 4.0)
        self.add_inward('tau', 5.0)
        self.add_inward('df_dt', 1.0)
        self.add_inward('max_step', 0.1)

        self.add_transient('f', der='df_dt',
            # max_time_step='min(0.01 * a / max(abs(df_dt), 1e-8), 0.2)',
            max_time_step='max_step * a / max(abs(df_dt), 1e-8)',
        )

    def make_case(self, driver_cls, **options):
        settings = options.copy()
        init = settings.pop('init', {})
        values = settings.pop('values', {})
        values['df_dt'] = self.df_dt_expr
        driver = self.add_driver(driver_cls(**settings))
        driver.set_scenario(values=values, init=init)
        return driver

    @property
    def df_dt_expr(self) -> str:
        """Expression of the derivative of f(t)"""
        return "-a * (t - t0) / tau**2 * exp(-0.5 * ((t - t0) / tau)**2)"

    def solution(self, t, init=(0, 0)) -> float:
        """Analytical solution at time t, with initial condition f(t0) = f0,
        where (t0, f0) is given by tuple `init`."""
        exact = lambda t: self.a * np.exp(-0.5 * ((t - self.t0) / self.tau)**2)
        t0, f0 = init
        return exact(t) - exact(t0) + f0


@pytest.fixture(scope='function')
def gaussian_ode():
    return GaussianOde('gaussian_ode')

# <codecell>

class ScalarOde(System):
    """System representing ODE df/dt = F(t)"""
    def setup(self):
        self.add_inward('df')
        self.add_transient('f', der='df')


class VectorOde(System):
    """System representing ODE dv/dt = V(t) in vectorial form"""
    def setup(self, size=3):
        self.add_inward('dv', np.zeros(max(size, 1)))
        self.add_transient('v', der='dv')


@pytest.fixture(scope='function')
def scalar_ode_case():
    return case_factory(ScalarOde, 'scalar_ode')

@pytest.fixture(scope='function')
def vector_ode_case():
    return case_factory(VectorOde, 'vector_ode')

# <codecell>

class FloatPort(Port):
    def setup(self):
        self.add_variable('value', 0.0)


class Tank(System):
    def setup(self, rho=1e3):
        self.add_inward('area', 1.0, desc='Cross-section area')
        self.add_inward('rho', abs(rho), desc='Fluid density')

        self.add_input(FloatPort, 'flowrate')
        self.add_output(FloatPort, 'p_bottom')

        self.add_transient('height', der='flowrate.value / area')

    def compute(self):
        g = 9.81
        self.p_bottom.value = self.rho * g * self.height


class Pipe(System):
    """Poiseuille flow in a cylindrical pipe"""
    def setup(self):
        self.add_inward('D', 0.1, desc="Diameter")
        self.add_inward('L', 2.0, desc="Length")
        self.add_inward('mu', 1e-3, desc="Fluid dynamic viscosity")

        self.add_input(FloatPort, 'p1')
        self.add_input(FloatPort, 'p2')

        self.add_output(FloatPort, 'Q1')
        self.add_output(FloatPort, 'Q2')

        self.add_outward('k', desc='Pressure loss coefficient')

    def compute(self):
        self.k = np.pi * self.D**4 / (256 * self.mu * self.L)
        self.Q1.value = self.k * (self.p2.value - self.p1.value)
        self.Q2.value = -self.Q1.value


class CoupledTanks(System):
    """System describing two tanks connected by a pipe (viscous limit)"""
    def setup(self, rho=1e3):
        self.add_child(Tank('tank1', rho=rho))
        self.add_child(Tank('tank2', rho=rho))
        self.add_child(Pipe('pipe'))

        self.connect(self.tank1.p_bottom, self.pipe.p1)
        self.connect(self.tank2.p_bottom, self.pipe.p2)
        self.connect(self.tank1.flowrate, self.pipe.Q1)
        self.connect(self.tank2.flowrate, self.pipe.Q2)


class CoupledTanksSolution:
    """Analytical solution of dynamic system CoupledTanks"""
    def __init__(self, system, init):
        K = 9.81 * system.tank1.rho * system.pipe.k
        a = system.tank1.area / system.tank2.area
        self.__area_ratio = a
        self.__tau = system.tank1.area / ((1 + a) * K)
        self.initial_heights = init

    @property
    def characteristic_time(self):
        return self.__tau

    def __call__(self, t):
        tau = self.__tau
        a = self.__area_ratio
        h1_0, h2_0 = self.initial_heights
        dh = (h1_0 - h2_0) * np.exp(-t / tau)
        h1 = (a * h1_0 + h2_0 + dh) / (1 + a)
        h2 = h1 - dh
        return (h1, h2)


@pytest.fixture(scope='function')
def two_tank_case():
    def _test_objects(driver_cls, rho=1e3, **kwargs):
        system = CoupledTanks('coupledTanks', rho=rho)
        driver = system.add_driver(driver_cls(**kwargs))
        return system, driver
    return _test_objects


@pytest.fixture(scope='function')
def two_tank_solution():
    def _test_objects(system, init):
        return CoupledTanksSolution(system, init)
    return _test_objects

# <codecell>

class PointMass(System):
    """Free fall of a point mass, with friction"""
    def setup(self):
        self.add_inward('mass', 1.2, desc='Mass')
        self.add_inward('k', 0.1, desc='Friction coefficient')
        self.add_inward('g', np.r_[0, 0, -9.81], desc='Uniform acceleration field')

        self.add_outward('a', np.zeros(3), desc='Acceleration')

        self.add_transient('v', der='a', desc='Velocity')
        self.add_transient('x', der='v', desc='Position', max_time_step='0.1 * norm(a)')
   
    def compute(self):
        self.a = self.g - (self.k / self.mass) * self.v


class PointMassSolution:
    """Analytical solution of dynamic system PointMass"""
    def __init__(self, system, v0, x0=np.zeros(3)):
        m = system.mass
        k = system.k
        g = np.array(system.g)
        if m <= 0:
            raise ValueError("Mass must be strictly positive")
        x0 = np.asarray(x0)
        v0 = np.asarray(v0)
        if k > 0:
            omega = k / m
            tau = 1 / omega
            A = g * tau
        else:
            omega, tau = 0, np.inf
            A = np.full_like(g, np.inf)
        B = v0 - A
        def x_solution(t):
            wt = omega * t
            if wt < 1e-7:  # asymptotic expansion, to avoid exp overflow
                x = x0 + v0 * t + (0.5 * t) * (g * t - wt * v0) * (1 - wt / 3 * (1 - 0.25 * wt))
            else:
                x = x0 + A * t + B * tau * (1 - np.exp(-wt))
            return x
        def v_solution(t):
            wt = omega * t
            if wt < 1e-7:  # asymptotic expansion, to avoid exp overflow
                v = v0 + (g * t - v0  * wt) * (1 - wt * (0.5 - wt / 6))
            else:
                v = A + B * np.exp(-wt)
            return v
        self.__x = x_solution
        self.__v = v_solution
        self.__a = lambda t: g - self.__v(t) * omega
        self.__omega = omega

    @property
    def omega(self):
        return self.__omega

    def a(self, t):
        return self.__a(t)

    def v(self, t):
        return self.__v(t)

    def x(self, t):
        return self.__x(t)


@pytest.fixture(scope='function')
def point_mass_case():
    return case_factory(PointMass, 'point')


@pytest.fixture(scope='function')
def point_mass_solution():
    def _test_objects(system, v0, x0=np.zeros(3)):
        return PointMassSolution(system, v0, x0)
    return _test_objects

# <codecell>

class PositionPort(Port):
    def setup(self):
        self.add_variable("x", np.zeros(3), desc="Position")

class KinematicsPort(Port):
    def setup(self):
        self.add_variable("v", np.zeros(3), desc="Velocity")

class DynamicsPort(Port):
    def setup(self):
        self.add_variable("a", np.zeros(3), desc="Acceleration")

class PointDynamics(System):
    def setup(self):
        self.add_inward("mass", 1.0)
        self.add_inward("acc_ext", np.zeros(3))
        self.add_inward("force_ext", np.zeros(3))

        self.add_outward("force", np.zeros(3))
        self.add_outward("acc", np.zeros(3))

    def compute(self):
        self.force = self.force_ext + self.mass * self.acc_ext
        self.acc = self.force / self.mass

class PointFriction(System):
    def setup(self):
        self.add_input(KinematicsPort, 'kinematics')
        self.add_inward("k", 0.1, desc="Friction coefficient")

        self.add_outward("force", np.zeros(3))

    def compute(self):
        self.force = -self.k * self.kinematics.v

class PointMassWithPorts(System):
    def setup(self):
        self.add_child(PointFriction('friction'), pulling=['k', 'kinematics'])
        self.add_child(PointDynamics('dynamics'), pulling={
            'mass': 'mass',
            'force': 'force',
            'acc_ext': 'g',
            'acc': 'a',
            })

        self.exec_order = ['friction', 'dynamics']

        self.add_input(PositionPort, 'position')

        self.connect(self.friction.outwards, self.dynamics.inwards, {"force": "force_ext"})

        self.g = np.r_[0, 0, -9.81]

        self.add_transient('kinematics.v', der='a')
        self.add_transient('position.x', der='kinematics.v', max_time_step='0.1 * norm(a)')


@pytest.fixture(scope='function')
def pointMassWithPorts_case():
    return case_factory(PointMassWithPorts, 'point')


class DevilCase(System):

    def setup(self):
        self.add_inward("x", np.zeros(3))
        self.add_inward("v", np.zeros(3))
        self.add_inward("a", np.zeros(3))

        self.add_inward("y", np.zeros(3))
        self.add_inward("w", np.zeros(3))

        self.add_inward("alpha", 0.0)
        self.add_inward("omega", 0.0)
        self.add_inward("zeta", 0.0)

        self.add_transient("x", der="v")
        self.add_transient("v", der="a")

        self.add_transient("y", der="w")
        self.add_transient("w", der="25. * a * zeta")

        self.add_transient("alpha", der="omega")
        self.add_transient("omega", der="zeta")

        self.add_transient("beta", der="42. * cos(zeta)")

# <codecell>

class SystemWithRate(System):
    def setup(self):
        self.add_inward('k', 2.0)
        self.add_inward('U')
        self.add_rate('dU_dt', source='U', initial_value='k')


@pytest.fixture(scope='function')
def rate_case_1():
    return case_factory(SystemWithRate, 'system')

# <codecell>

class DampedMassSpring(System):
    """Harmonic oscillator"""
    def setup(self):
        self.add_inward('mass', 0.25)
        self.add_inward('length', 0.5, desc='Unloaded spring length')
        self.add_inward('c', 0.0, desc='Friction coefficient')
        self.add_inward('K', 0.1, desc='Spring stiffness')
        self.add_inward('x', 0.0, desc='Linear position')

        self.add_outward('F', 0.0, desc='Force')
        self.add_outward('a', 0.0, desc='Linear acceleration')

        self.add_transient('v', der='a')
        self.add_transient('x', der='v', max_time_step='0.1 * sqrt(mass / K)')
        
    def compute(self):
        self.F = self.K * (self.length - self.x) - self.c * self.v
        self.a = self.F / self.mass


class HarmonicOscillatorSolution:
    def __init__(self, system, x0, v0=0):
        K = system.K
        m = system.mass
        L = system.length
        x0 -= L
        w0 = np.sqrt(K / m)
        self.__damping = dc = 0.5 * system.c / np.sqrt(m * K)
        wd = w0 * dc
        if self.over_damped:
            ws = w0 * np.sqrt(dc**2 - 1)
            A, B = 0.5 * (v0 + (wd + ws) * x0) / ws, 0.5 * (v0 + (wd - ws) * x0) / ws
            self.__x = lambda t: L + A * np.exp(-(wd - ws) * t) - B * np.exp(-(wd + ws) * t)
        else:
            ws = w0 * np.sqrt(1 - dc**2)
            A, B = (v0 + wd * x0) / ws, x0
            self.__x = lambda t: L + np.exp(-wd * t) * (A * np.sin(ws * t) + B * np.cos(ws * t))
        
    def x(self, t):
        return self.__x(t)

    @property
    def damping(self):
        """Damping coefficient"""
        return self.__damping
    
    @property
    def over_damped(self):
        return self.__damping > 1


@pytest.fixture(scope='function')
def oscillator_case():
    return case_factory(DampedMassSpring, 'system')


@pytest.fixture(scope='function')
def oscillator_solution():
    def _test_objects(system, x0, v0=0):
        return HarmonicOscillatorSolution(system, x0, v0)
    return _test_objects



import numpy


class PointMassSolution:
    """Analytical solution of dynamic system PointMass"""
    def __init__(self, system, v0, x0=numpy.zeros(3)):
        if system.mass <= 0:
            raise ValueError("Mass must be strictly positive")
        x0 = numpy.asarray(x0)
        v0 = numpy.asarray(v0)
        g = numpy.asarray(system.g)
        if system.k > 0:
            omega = system.k / system.mass
            tau = 1 / omega
            A = g * tau
        else:
            omega, tau = 0, numpy.inf
            A = numpy.full_like(g, numpy.inf)
        B = v0 - A
        def x_impl(t):
            wt = omega * t
            if wt < 1e-7:  # asymptotic expansion, to avoid exp overflow
                x = x0 + v0 * t + (0.5 * t) * (g * t - wt * v0) * (1 - wt / 3 * (1 - 0.25 * wt))
            else:
                x = x0 + A * t + B * tau * (1 - numpy.exp(-wt))
            return x
        def v_impl(t):
            wt = omega * t
            if wt < 1e-7:  # asymptotic expansion, to avoid exp overflow
                v = v0 + (g * t - v0  * wt) * (1 - wt * (0.5 - wt / 6))
            else:
                v = A + B * numpy.exp(-wt)
            return v
        self.__x = x_impl
        self.__v = v_impl
        self.__a = lambda t: g - self.__v(t) * omega
        self.__omega = omega

    @property
    def omega(self):
        return self.__omega

    def a(self, t):
        return self.__a(t)

    def v(self, t):
        return self.__v(t)

    def x(self, t):
        return self.__x(t)


class CoupledTanksSolution:
    """Analytical solution of dynamic system CoupledTanks"""
    def __init__(self, system, init):
        K = 9.81 * system.tank1.rho * system.pipe.k
        a = system.tank1.area / system.tank2.area
        self.__area_ratio = a
        self.__tau = system.tank1.area / ((1 + a) * K)
        self.initial_heights = init

    @property
    def characteristic_time(self):
        return self.__tau

    def __call__(self, t):
        tau = self.__tau
        a = self.__area_ratio
        h1_0, h2_0 = self.initial_heights
        dh = (h1_0 - h2_0) * numpy.exp(-t / tau)
        h1 = (a * h1_0 + h2_0 + dh) / (1 + a)
        h2 = h1 - dh
        return (h1, h2)


class HarmonicOscillatorSolution:
    def __init__(self, system, init=(0, 0)):
        K = system.K
        c = system.c
        m = system.mass
        L = system.length
        w0 = numpy.sqrt(K / m)
        self.__damping = dc = 0.5 * c / numpy.sqrt(m * K)
        a = w0 * dc
        x0, v0 = init
        x0 -= L
        if self.over_damped:
            wd = w0 * numpy.sqrt(dc**2 - 1)
            A, B = 0.5 * (v0 + (a + wd) * x0) / wd, 0.5 * (v0 + (a - wd) * x0) / wd
            self.__x = lambda t: L + (A * numpy.exp(-(a - wd) * t) - B * numpy.exp(-(a + wd) * t))
        else:
            wd = w0 * numpy.sqrt(1 - dc**2)
            A, B = (v0 + a * x0) / wd, x0
            self.__x = lambda t: L + numpy.exp(-a * t) * (A * numpy.sin(wd * t) + B * numpy.cos(wd * t))
        
    def x(self, t):
        return self.__x(t)

    @property
    def damping(self):
        """Damping coefficient"""
        return self.__damping
    
    @property
    def over_damped(self):
        return self.__damping > 1


from multiprocessing import Process, Queue
from typing import Dict, Any

from cosapp.systems.externalsystem import ExternalSystem


class ProcessSystem(ExternalSystem):

    __slots__ = ('_queue_in', '_queue_out', 'to_execute')

    def __init__(self, name: str, **kwargs):
        self._queue_in = Queue()
        self._queue_out = Queue()
        self.to_execute = Queue()

        super().__init__(name, **kwargs)

    def _initialize(self, **kwargs):
        self.add_inward(
            "working_folder",
            None,
            dtype=str,
            desc="Folder of execution for the process.",
        )
        return kwargs

    def call_setup_run(self):
        super().call_setup_run()
        self._process = Process(
            target=self.to_execute,
            args=(self._queue_in, self._queue_out, self.working_folder),
        )
        self._process.start()

    def call_clean_run(self):
        self._queue_in.put({})
        # self._queue_out.get()
        self._process.join()
        super().call_clean_run()

    def send_inputs(self):
        self._queue_in.put(self.serialize_data())

    def read_outputs(self, timeout: float = 1) -> Dict[str, Any]:
        return self._queue_out.get(True, timeout)



import pytest

import numpy as np

from cosapp.core.numerics.boundary import TimeUnknown
from cosapp.core.eval_str import EvalString
from cosapp.systems import System
from cosapp.drivers.time.utils import TimeUnknownDict, TimeStepManager


class SubSystem(System):
    def setup(self):
        self.add_inward('x', 1.0)
        self.add_inward('y', np.ones(3))


class DummySystem(System):
    def setup(self):
        self.add_child(SubSystem('sub1'))
        self.add_child(SubSystem('sub2'))

        self.add_inward('f', 0.0)


@pytest.fixture(scope="function")
def system():
    return DummySystem("system")


@pytest.fixture(scope="function")
def unknowns(system):
    """Returns a list of dummy unknowns"""
    a = TimeUnknown(system, "f", der=0.5)
    b = TimeUnknown(system, "sub1.y", der="sub1.x")
    c = TimeUnknown(system, "sub2.x", der="sub1.y**2", max_time_step="0.1 * sub1.x")
    return a, b, c


@pytest.fixture(scope="function")
def unknown_dict(unknowns):
    a, b, c = unknowns
    return TimeUnknownDict(a=a, b=b, c=c)


@pytest.fixture(scope="function")
def manager(unknown_dict):
    return TimeStepManager(unknown_dict)


@pytest.mark.parametrize("data, expected", [
    (dict(), dict()),
    (dict(use_unknowns=True), dict()),
    (dict(use_unknowns=True, dt=None, max_growth_rate=2), dict(dt=None, max_growth_rate=2)),
    (dict(dt=0.01, max_growth_rate=1.7), dict(dt=0.01, max_growth_rate=1.7)),
    (dict(dt=0, max_growth_rate=1.5), dict(error=ValueError, mach="dt")),
    (dict(max_growth_rate=0.5), dict(error=ValueError, mach="max_growth_rate")),
    (dict(max_growth_rate=1), dict(error=ValueError, mach="max_growth_rate")),
    (dict(max_growth_rate=1), dict(error=ValueError, mach="max_growth_rate")),
    (dict(max_growth_rate=-0.2), dict(error=ValueError, mach="max_growth_rate")),
])
def test_TimeStepManager__init__(unknown_dict, data, expected):
    if data.pop('use_unknowns', False):  # use unknown_dict fixture?
        data.setdefault('transients', unknown_dict)
        expected['length'] = len(unknown_dict)
    else:
        expected.setdefault('length', 0)

    error = expected.get('error', None)

    if error is None:
        manager = TimeStepManager(**data)
        assert isinstance(manager.transients, TimeUnknownDict)
        assert len(manager.transients) == expected['length']
        assert manager.max_growth_rate > 1
        assert manager.max_growth_rate == expected.get('max_growth_rate', np.inf)
        assert manager.nominal_dt == expected.get('dt', None)
    else:
        pattern = expected.get('match', None)
        with pytest.raises(error, match=pattern):
            TimeStepManager(**data)


@pytest.mark.parametrize("options, expected", [
    (dict(), dict(dt=0.3)),
    (dict(dt=0.1), dict(dt=0.1)),
    (dict(dt=0.01), dict(dt=0.01)),
])
def test_TimeStepManager_time_step(unknown_dict, options, expected):
    manager = TimeStepManager(unknown_dict, **options)

    s = manager.transients["a"].context
    s.sub1.x = 3
    s.sub1.y[:] = 1.2
    s.sub2.x = 0
    s.sub2.y[:] = 0

    assert manager.time_step() == pytest.approx(expected['dt'], rel=1e-14)


@pytest.mark.parametrize("options, prev_dt, expected", [
    (dict(dt=0.1), None, 0.1),
    (dict(dt=0.1), 0.01, 0.1),
    (dict(dt=0.1, max_growth_rate=2), None, 0.1),
    (dict(dt=0.1, max_growth_rate=2), 0.01, 0.02),
    (dict(dt=0.1, max_growth_rate=1.5), 0.01, 0.015),
    # Check that prev_dt is disregarded when prev_dt <= 0 or dt < prev_dt:
    (dict(dt=0.1, max_growth_rate=2), 0.5, 0.1),
    (dict(dt=0.1, max_growth_rate=2), -10, 0.1),
    (dict(dt=0.1, max_growth_rate=2), 0.0, 0.1),
])
def test_TimeStepManager_time_step_limited(unknowns, options, prev_dt, expected):
    x = unknowns[1]
    manager = TimeStepManager(TimeUnknownDict(x=x), **options)

    assert manager.time_step(prev_dt) == pytest.approx(expected, rel=1e-14)


@pytest.mark.parametrize("rate, expected", [
    (1.5, dict()),
    (2, dict()),
    (1 + 1e-6, dict(error=None)),
    # Erroneous values
    ("1.5", dict(error=TypeError)),
    (0, dict(error=ValueError)),
    (0.5, dict(error=ValueError)),
    (1, dict(error=ValueError)),
    (-1, dict(error=ValueError)),
    (-0.2, dict(error=ValueError)),
])
def test_TimeStepManager_max_growth_rate(manager, rate, expected):
    error = expected.get('error', None)

    if error is None:
        manager.max_growth_rate = rate
        assert manager.max_growth_rate == expected.get('rate', rate)
    else:
        with pytest.raises(error):
            manager.max_growth_rate = rate



# -*- coding: utf-8 -*-
"""The setup script."""
import os
from typing import Iterable, Union, List

from setuptools import setup
from shutil import copytree, rmtree

# Copy the tutorials in cosapp, so they are included in the wheel
def to_ignore(
    folder: Union[bytes, str], filenames: List[Union[bytes, str]]
) -> Iterable[Union[bytes, str]]:
    ignored_files = list()
    for name in filenames:
        if name in (".ipynb_checkpoints",):  # List of subfolders to ignore
            ignored_files.append(name)

        root, ext = os.path.splitext(name)
        if ext not in (
            "",
            ".ipynb",
            ".py",
            ".svg",
        ):  # List of extension to keep - empty ext == subfolder
            ignored_files.append(name)

    return ignored_files


if os.path.exists("cosapp/tutorials"):
    rmtree("cosapp/tutorials")
copytree("docs/tutorials", "cosapp/tutorials", ignore=to_ignore)

# See setup.cfg for all options and MANIFEST.in for data files
setup()

if os.path.exists("cosapp/tutorials"):
    rmtree("cosapp/tutorials")



"""
Module handling the execution of code provided as string by the user.

This code is inspired from the OpenMDAO module openmdao.components.exec_comp.
"""
import numpy
from numbers import Number
from typing import Any, Dict, Iterable, Optional, Tuple, Union, Callable

from cosapp.utils.helpers import check_arg


class ContextLocals(dict):
    """Sub-set of context attributes.
    
    Parameters
    ----------
    context: System
        System whose attributes are looked up
    """

    def __init__(self, context: "System", *args, **kwargs):
        super().__init__(*args, *kwargs)
        self.__context = context
    
    @property
    def context(self) -> "System":
        """cosapp.systems.System: Context of the locals"""
        return self.__context

    def __missing__(self, key: Any) -> Any:
        try:
            value = getattr(self.__context, key)
        except AttributeError:
            raise KeyError(key)
        else:
            self[key] = value
            return value


class EvalString:
    """Create a executable statement using an expression string.

    The following functions are available for use in expression:

    =========================  ====================================
    Function                   Description
    =========================  ====================================
    abs(x)                     Absolute value of x
    acos(x)                    Inverse cosine of x
    acosh(x)                   Inverse hyperbolic cosine of x
    arange(start, stop, step)  Array creation
    arccos(x)                  Inverse cosine of x
    arccosh(x)                 Inverse hyperbolic cosine of x
    arcsin(x)                  Inverse sine of x
    arcsinh(x)                 Inverse hyperbolic sine of x
    arctan(x)                  Inverse tangent of x
    asin(x)                    Inverse sine of x
    asinh(x)                   Inverse hyperbolic sine of x
    atan(x)                    Inverse tangent of x
    cos(x)                     Cosine of x
    cosh(x)                    Hyperbolic cosine of x
    cross(x, y)                Cross product of arrays x and y
    dot(x, y)                  Dot-product of x and y
    e                          Euler's number
    erf(x)                     Error function
    erfc(x)                    Complementary error function
    exp(x)                     Exponential function
    expm1(x)                   exp(x) - 1
    factorial(x)               Factorial of all numbers in x
    fmax(x, y)                 Element-wise maximum of x and y
    fmin(x, y)                 Element-wise minimum of x and y
    inner(x, y)                Inner product of arrays x and y
    isinf(x)                   Element-wise detection of numpy.inf
    isnan(x)                   Element-wise detection of numpy.nan
    kron(x, y)                 Kronecker product of arrays x and y
    linspace(x, y, N)          Numpy linear spaced array creation
    log(x)                     Natural logarithm of x
    log10(x)                   Base-10 logarithm of x
    log1p(x)                   log(1+x)
    matmul(x, y)               Matrix multiplication of x and y
    maximum(x, y)              Element-wise maximum of x and y
    minimum(x, y)              Element-wise minimum of x and y
    ones(N)                    Create an array of ones
    outer(x, y)                Outer product of x and y
    pi                         Pi
    power(x, y)                Element-wise x**y
    prod(x)                    The product of all elements in x
    sin(x)                     Sine of x
    sinh(x)                    Hyperbolic sine of x
    sum(x)                     The sum of all elements in x
    round(x)                   Round all elements in x
    tan(x)                     Tangent of x
    tanh(x)                    Hyperbolic tangent of x
    tensordot(x, y)            Tensor dot product of x and y
    zeros(N)                   Create an array of zeros
    =========================  ====================================

    Full list is returned by `EvalString.available_symbols()`

    Parameters
    ----------
    expression : str
        Interpretable Python statement. In addition to standard Python operators,
        a subset of numpy and scipy functions is supported.
    context : cosapp.core.module.Module or cosapp.core.numerics.basics.Residue
        System or Residue defining the local context in which the statement will be executed

    Notes
    -----
    The context is used to include the `System` ports, children, inwards and outwards.
    If the context is a `Residue`, the reference value from the residue is added as
    `residue_reference` variable in the execution context.
    """

    # this dict will act as the global scope when we eval our expressions
    __globals = {}  # type: Dict[str, Any]

    @classmethod
    def available_symbols(cls) -> Dict[str, Any]:
        """
        List of available symbols (constants and functions) in current execution context.

        Returns
        -------
        Dict[str, Any]
            Mapping of available symbols by their name.
        """
        mapping = cls.__globals

        if len(mapping) > 0:
            return mapping
        
        def add_symbols(
            module: object, names: Optional[Iterable[str]] = None
        ) -> None:
            """
            Map attribute names from the given module into the global dict.

            Parameters
            ----------
            mod : object
                Module to check.
            names : iter of str, optional
                If supplied, only map attrs that match the given names
            """
            if names is None:
                names = dir(module)
            for name in names:
                if isinstance(name, tuple):
                    name, alias = name
                else:
                    alias = name
                if not name.startswith("_"):
                    mapping[alias] = mapping[name] = getattr(module, name)

        add_symbols(numpy,
            names=[
                # Numpy types
                "int8",
                "int16",
                "int32",
                "int64",
                "uint8",
                "uint16",
                "uint32",
                "uint64",
                "float32",
                "float64",
                "complex64",
                "complex128",
                # Array functions
                "abs",
                "array",
                "asarray",
                "arange",
                "concatenate",
                "ones",
                "zeros",
                "full",
                "full_like",
                "linspace",  # Array creation
                "e",
                "pi",  # Constants
                "inf",
                "isinf",
                "isnan",  # Logic
                "log",
                "log10",
                "log1p",
                "power",
                "sqrt",  # Math operations
                "cbrt",
                "exp",
                "expm1",
                "fmax",
                "fmin",
                "maximum",
                "minimum",
                "round",
                "sum",
                "dot",
                "prod",  # Reductions
                "tensordot",
                "matmul",  # Linear algebra
                "cross",
                "outer",
                "inner",
                "kron",
                "sin",
                "cos",
                "tan",
                ("arcsin", "asin"),  # Trig
                ("arccos", "acos"),
                ("arctan", "atan"),
                "sinh",
                "cosh",
                "tanh",
                ("arcsinh", "asinh"),  # Hyperbolic trig
                ("arccosh", "acosh"),
            ],
        )
        add_symbols(numpy.linalg, names=["norm"])

        # if scipy is available, add few specials functions
        try:
            import scipy.special
        except ImportError:
            pass
        else:
            add_symbols(scipy.special,
                names=["factorial", "erf", "erfc"]
            )

        # Add residues helper function
        from cosapp.core.numerics.residues import Residue
        mapping["evaluate_residue"] = Residue.evaluate_residue
        mapping["residue_norm"] = Residue.residue_norm

        return mapping

    def __init__(self, expression: Any, context: "System") -> None:
        """Class constructor.

        Compiles an expression, and checks that it is evaluable within a given context.
        """
        from cosapp.systems import System
        if not isinstance(context, System):
            cname = type(context).__name__
            raise TypeError(
                f"Object of type {cname!r} is not a valid context to evaluate expression '{expression}'."
            )

        self.__str = EvalString.string(expression)  # type: str
        if len(self.__str) == 0:
            raise ValueError("Can't evaluate empty expressions")

        code = compile(self.__str, "<string>", "eval")  # type: CodeType

        # Look for the requested variables
        global_dict = self.available_symbols()
        self.__locals = local_dict = ContextLocals(context)  # type: ContextLocals
        value = eval(code, global_dict, local_dict)
        
        constants = context.properties
        self.__constant = set(local_dict).issubset(constants)  # type: bool

        if self.__constant:
            # simply return constant value
            eval_impl = lambda: value
        else:            
            # By specifying global and local contexts, we limit the user scope.
            eval_impl = lambda: eval(code, self.globals, self.locals)
        self.__eval = eval_impl  # type: Callable[[], Any]

    @staticmethod
    def string(expression: Any) -> str:
        """Converts an expression into a suitably formatted string.

        Notes
        -----
        Necessary step for numpy arrays (and possibly other types), whose plain string
        conversion is not readily evaluable (hence the use of 'repr' instead of 'str').

        Examples
        --------
        >>> str(numpy.array([0.1, 0.2]))
        [0.1 0.2]
        
        >>> repr(numpy.array([0.1, 0.2]))
        array([0.1, 0.2])
        """
        if isinstance(expression, str):
            return expression.strip()
        elif isinstance(expression, EvalString):
            return str(expression)
        else:
            return repr(expression)

    def __str__(self) -> str:
        return self.__str

    def __repr__(self) -> str:
        return repr(self.__str)

    def __contains__(self, pattern: str):
        return (pattern in self.__str)

    @property
    def locals(self) -> Dict[str, Any]:
        """Dict[str, Any]: Context attributes required to evaluate the string expression."""
        # Read attribute values from context object
        eval_context = self.eval_context
        for key in self.__locals:
            self.__locals[key] = getattr(eval_context, key)
        return self.__locals

    @property
    def globals(self) -> Dict[str, Any]:
        """Dict[str, Any]: Global functions and variables required to evaluate the string expression."""
        return EvalString.__globals

    @property
    def eval_context(self) -> "System":
        """cosapp.systems.System: Context of string expression evaluation."""
        return self.__locals.context

    @property
    def constant(self) -> bool:
        """
        Returns a boolean indicating whether or not the evaluated expression
        is constant, that is independent of its context.
        """
        return self.__constant

    def eval(self) -> Any:
        """Evaluate the expression in the system context.

        Returns
        -------
        Any
            The result of the expression evaluation.
        """
        return self.__eval()


class AssignString:
    """Create an executable assignment of the kind 'lhs = rhs' from two evaluable expressions lhs and rhs."""

    def __init__(self, lhs: str, rhs: Any, context: "System") -> None:
        lhs = EvalString(lhs, context)
        if lhs.constant:
            raise ValueError(
                f"The left-hand side of an assignment expression cannot be constant ({lhs!r})")
        # At this point, lhs is a valid expression within given context
        self.__sides = None
        self.__raw_sides = [str(lhs), str(rhs)]  # raw sides lhs and rhs, without reformatting
        value = lhs.eval()
        if isinstance(value, numpy.ndarray):
            self.__shape = value.shape
            self.__dtype = value.dtype
        else:
            self.__shape = None
            self.__dtype = type(value)
        self.__context = context
        self.__locals = lhs.locals
        self.__locals.update({"rhs_value": value, context.name: context})
        assignment = f"{context.name}.{lhs!s} = rhs_value"
        self.__code = compile(assignment, "<string>", "exec")  # assignment bytecode
        self.rhs = rhs

    @property
    def eval_context(self) -> "System":
        """Evaluation context of the assignment."""
        return self.__context

    @property
    def lhs(self) -> str:
        """Left-hand side of the assignment, as a string"""
        return self.__raw_sides[0]

    @property
    def rhs(self) -> str:
        """Right-hand side of the assignment, as a string"""
        return self.__raw_sides[1]

    @rhs.setter
    def rhs(self, rhs: Any) -> None:
        context = self.eval_context
        lhs = self.lhs
        erhs = EvalString(rhs, context)
        raw_rhs = str(erhs)
        value = erhs.eval()
        if erhs.constant:
            rhs = EvalString.string(value)
        else:
            rhs = raw_rhs
        sides = EvalString(f"({lhs}, {rhs})", context)
        if self.__shape:
            if value is None:
                value = sides.eval()[1]
            if isinstance(value, Number):
                sides = EvalString(f"({lhs}, full_like({lhs}, {rhs}))", context)
            else:  # tentatively copy rhs into a numpy array
                self.__check_size(value)
                sides = EvalString(f"({lhs}, array({rhs}, dtype={self.__dtype}))", context)
        self.__sides = sides
        self.__constant = erhs.constant
        self.__raw_sides[1] = raw_rhs

    @property
    def constant(self) -> bool:
        """Bool: is the right-hand side of the assignment constant?"""
        return self.__constant

    def exec(self, context=None) -> Tuple[Any, bool]:
        """
        Evaluates rhs, and executes assignment lhs <- rhs.

        Returns
        -------
        Tuple[Any, bool]
            (rhs, changed), where 'changed' is True if the value of rhs has changed, False otherwise.
        """
        sides = self.__sides.eval()  # updates context at the same time
        changed = not numpy.array_equal(sides[0], sides[1])
        if context is None:
            context = self.__locals
        context['rhs_value'] = sides[1]
        exec(self.__code, dict(), context)
        return sides[1], changed

    @property
    def shape(self) -> Union[Tuple[int, int], None]:
        """Returns the shape of the assigned object (lhs) if it is an array, else None."""
        return self.__shape

    def __str__(self) -> str:
        return " = ".join(self.__raw_sides)

    def __repr__(self) -> str:
        cls_name = self.__class__.__qualname__
        return "{}({!r}, {!r}, {})".format(cls_name, *self.__raw_sides, self.__context.name)

    def __check_size(self, array):
        """Checks if `array` is shape-compatible with lhs."""
        shape = self.__shape
        if numpy.shape(array) != shape:
            raise ValueError(f"Cannot assign {array} to array of shape {shape}")



"""
Classes containing all `System` variables.
"""
import array
import logging
from collections import OrderedDict
from collections.abc import MutableSequence
from typing import Any, Dict, Iterator, Optional, Tuple, Union
from types import MappingProxyType
import numpy

from cosapp.core.numerics.distributions.distribution import Distribution
from cosapp.ports.enum import PortType, Scope, Validity
from cosapp.ports.exceptions import ScopeError
from cosapp.ports.variable import RangeValue, Types, Variable
from cosapp.utils.helpers import check_arg
from cosapp.utils.naming import NameChecker

logger = logging.getLogger(__name__)


class ExtensiblePort:
    """`ExtensiblePort` is a container to gather variables.

    Common user should not use this class.

    Parameters
    ----------
    name : str
        Port name
    direction : {PortType.IN, PortType.OUT}
        Port direction
    """

    _name_check = NameChecker()

    def __init__(self, name: str, direction: PortType) -> None:
        """`ExtensiblePort` constructor.

        Parameters
        ----------
        name : str
            Port name
        direction : {PortType.IN, PortType.OUT}
            Port direction
        """
        if not isinstance(direction, PortType):
            raise TypeError(f"Direction must be PortType; got {direction}.")

        self._variables = OrderedDict()  # type: Dict[str, Variable]
        self._name = self._name_check(name)  # type: str
        self._direction = direction  # type: PortType
        self._owner = None  # type: Optional[cosapp.systems.System]
        self.__clearance = None
        self.scope_clearance = Scope.PRIVATE

    @property
    def owner(self) -> "Optional[cosapp.systems.System]":
        """System : `System` owning the port."""
        return self._owner

    @owner.setter
    def owner(self, new_owner: "cosapp.systems.System"):
        from cosapp.systems import System

        if not isinstance(new_owner, System):
            raise TypeError(f"Port owner must be a `System`; got {new_owner!r}.")
        self._owner = new_owner
        self.__update_filter()

    @property
    def name(self) -> str:
        """str : Port name"""
        return self._name

    @property
    def contextual_name(self) -> str:
        """str : Join port owner name and port name.

        If the port has no owner, the port name is returned.
        """
        owner = self._owner
        return self._name if owner is None else f"{owner.name}.{self._name}"

    def full_name(self, trim_root=False) -> str:
        """Returns full name up to root owner.
        
        Parameters
        ----------
        trim_root : bool (optional, default False)
            Exclude root owner name if True.

        Returns
        -------
        str
            The port full name
        """
        owner = self.owner
        path = []
        if owner is not None:
            path = owner.path_namelist()
            if trim_root:
                path = path[1:]
        path.append(self.name)
        return ".".join(path)

    @property
    def direction(self) -> PortType:
        """:obj:`PortType.IN` or :obj:`PortType.OUT` : Port direction"""
        return self._direction

    def __repr__(self) -> str:
        var_list = [f"'{key}': {getattr(self, key)!r}" for key in self._variables]
        return f"{type(self).__name__}: {{{', '.join(var_list)}}}"

    def _repr_markdown_(self) -> str:
        """Returns the representation of this port variables in Markdown format.

        Returns
        -------
        str
            Markdown formatted representation
        """
        from cosapp.tools.views.markdown import port_to_md

        return port_to_md(self)

    def __json__(self) -> Dict[str, Dict[str, Any]]:
        """JSONable dictionary representing a variable.
        
        Returns
        -------
        Dict[str, Any]
            The dictionary
        """
        rtn = dict()
        for name, var in self._variables.items():
            rtn[name] = var.__json__()

        return rtn

    def serialize_data(self) -> Dict[str, Any]:
        """Serialize the variable values in a dictionary.
        
        Returns
        -------
        Dict[str, Any]
            The dictionary (variable name, value)
        """
        rtn = dict()
        for name in self._variables:
            rtn[name] = getattr(self, name)
        return rtn

    def add_variable(
        self,
        name: str,
        value: Any = 1,
        unit: str = "",
        dtype: Types = None,
        valid_range: RangeValue = None,
        invalid_comment: str = "",
        limits: RangeValue = None,
        out_of_limits_comment: str = "",
        desc: str = "",
        distribution: Optional[Distribution] = None,
        scope: Scope = Scope.PRIVATE,
    ) -> None:
        """Add a variable to the port.

        The `valid_range` defines the range of value for which a model is known to behave correctly.
        The `limits` are at least as large as the validity range.

        Parameters
        ----------
        name : str
            Name of the variable
        value : Any, optional
            Value of the variable; default 1
        unit : str, optional
            Variable unit; default empty string (i.e. dimensionless)
        dtype : type or iterable of type, optional
            Variable type; default None (i.e. type of initial value)
        valid_range : Tuple[Any, Any] or Tuple[Tuple], optional
            Validity range of the variable; default None (i.e. all values are valid)
        invalid_comment : str, optional
            Comment to show in case the value is not valid; default ''
        limits : Tuple[Any, Any] or Tuple[Tuple], optional
            Limits over which the use of the model is wrong; default valid_range
        out_of_limits_comment : str, optional
            Comment to show in case the value is not valid; default ''
        desc : str, optional
            Variable description; default ''
        distribution : Distribution, optional
            Probability distribution of the variable; default None (no distribution)
        scope : Scope {PRIVATE, PROTECTED, PUBLIC}, optional
            Variable visibility; default PRIVATE
        """
        if name in self:
            logger.warning(
                f"Variable {name} already exists in port {self.contextual_name}."
                " It will be overwritten."
            )

        # Value must be set before creating variable as validation in Variable needs it
        self._variables[name] = Variable(
            name,
            self,
            value,
            unit,
            dtype,
            valid_range,
            invalid_comment,
            limits,
            out_of_limits_comment,
            desc,
            distribution,
            scope,
        )

        if self._variables[name].dtype == (
            MutableSequence,
            array.ArrayType,
            numpy.ndarray,
        ):
            value = numpy.asarray(value)

        # For efficiency reasons, variables are stored as port attributes
        setattr(self, name, value)

    # TODO unused and should be removed -> to dangerous for consistency
    def remove_variable(self, name: str) -> None:
        """Remove the variable from the port.

        Parameters
        ----------
        name : str
            Name of the variable to be removed

        Raises
        ------
        ValueError
            If the variable does not exists
        """

        if name not in self._variables:
            raise AttributeError(
                f"Variable {name!r} does not exist in port {self.contextual_name}"
            )

        delattr(self, name)
        self._variables.pop(name)

    # TODO should we override __setattr__ to forward setting to the source port? and/or raise an
    #   error if it is not possible - for example if the source is an output variable
    def __set_notype_checking(self, key, value):
        super().__setattr__(key, value)

        if key in self._variables and self._owner:
            self._owner.set_dirty(self.direction)

    def validate(self, key: str, value: Any) -> None:
        """Check if a variable is in the scope of the user and the type is valid.
        
        Parameters
        ----------
        key : str
            Name of the variable to test
        value : Any
            Value to validate

        Raises
        ------
        ScopeError
            If the variable is not visible for the user
        TypeError
            If the value has an unauthorized type
        """
        # Check scope
        if self.__out_of_scope(key):
            raise ScopeError(f"Cannot set out-of-scope variable {key!r}.")

        if value is not None:
            var_type = self._variables[key].dtype
            ok = var_type is None or isinstance(value, var_type)
            if not ok:
                raise TypeError(
                    "Trying to set {}.{} of type {} with {}.".format(
                        self.contextual_name, key, var_type, type(value)
                    )
                )

    def __set_variable(self, key, value):
        """Set the variable `key` with `value`.
        
        Parameters
        ----------
        key : str
            Name of the variable to be set
        value : Any
            Value to set
        """
        if key.startswith("_"):
            super().__setattr__(key, value)
        elif key in self._variables:
            self.validate(key, value)
            self.__set_notype_checking(key, value)
        elif hasattr(self, key):
            super().__setattr__(key, value)
        else:
            raise AttributeError(
                f"Port variable {self.contextual_name}.{key} can only be created using method 'add_variable'."
            )

    __setattr__ = __set_variable

    @staticmethod
    def set_type_checking(activate: bool) -> None:
        """(Un)set type checking when affecting port variables.

        By default type checking is activated.

        Parameters
        ----------
        activate : bool
            True to activate type checking, False to deactivate
        """
        if activate:
            ExtensiblePort.__setattr__ = ExtensiblePort.__set_variable
        else:
            ExtensiblePort.__setattr__ = ExtensiblePort.__set_notype_checking

    def __contains__(self, item: str) -> bool:
        return item in self._variables

    def __getitem__(self, item: str) -> Any:
        try:
            return getattr(self, item)
        except AttributeError:
            raise KeyError(f"Variable or property {item} does not exist in Port {self}.")

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self._variables:
            setattr(self, key, value)
        else:
            raise KeyError(f"Variable {key} does not exist in Port {self}.")

    def __iter__(self) -> Iterator[str]:
        return iter(self._variables)

    def __len__(self) -> int:
        return len(self._variables)

    @property
    def scope_clearance(self) -> Scope:
        """Scope: Current clearance level of the port.
        Determines the set of read-only port variables."""
        return self.__clearance

    @scope_clearance.setter
    def scope_clearance(self, user_scope: Scope) -> None:
        check_arg(user_scope, "scope_clearance", Scope)
        self.__clearance = user_scope
        self.__update_filter()

    def __update_filter(self) -> None:
        """Update the implementation of method `out_of_scope`
        according to port's clearance level and owner."""
        if self._owner is None:
            criterion = lambda name: False
        else:
            read_only = [
                n for n, details in self._variables.items() if details.scope > self.__clearance
            ]
            criterion = lambda name: not self._owner.is_running() and name in read_only
        self.__out_of_scope = criterion

    def out_of_scope(self, name: str) -> bool:
        """
        Asserts if current scope `scope_clearance` is high enough to
        allow the modification of port variable `name`.

        Parameters
        ----------
        name : str
            Variable name to test

        Returns
        -------
        bool
            Is modification forbidden?

        The behavior of this function is set by method `scope_clearance()`.
        By default, clearance level is set to `Scope.PUBLIC`.

        Examples
        --------
        >>> port = Port("myPort", PortType.IN)
        >>> port.add_variable("x", 1.5, scope=Scope.PUBLIC)
        >>> port.add_variable("y", 0.2, scope=Scope.PROTECTED)
        >>> port.add_variable("z", 0.3, scope=Scope.PRIVATE)
        >>>
        >>> port.scope_clearance(Scope.PROTECTED)  # only `port.x` and `port.y` can be modified
        >>> assert not port.out_of_scope("x")
        >>> assert not port.out_of_scope("y")
        >>> assert port.out_of_scope("z")
        >>>
        >>> port.scope_clearance(Scope.PUBLIC)     # only `port.x` can be modified
        >>> assert not port.out_of_scope("x")
        >>> assert port.out_of_scope("y")
        >>> assert port.out_of_scope("z")
        """
        return self.__out_of_scope(name)

    def get_details(self, name: Optional[str] = None) -> Union[Dict[str, Variable], Variable]:
        """Return the variable(s).
        
        Parameters
        ----------
        name : str, optional
            Name of the variable looked for; default None (all variable are returned).

        Returns
        -------
        types.MappingProxyType[str, cosapp.ports.variable.Variable] or cosapp.ports.variable.Variable
            The sought variable, or a read-only view on all port variables.
        """
        if name is None:
            return MappingProxyType(self._variables)
        else:
            return self._variables[name]

    def check(self, name: Optional[str] = None) -> Union[Dict[str, Validity], Validity]:
        """Get the variable value validity.

        If `name` is not provided, returns a dictionary with the validity of all variables. Else
        only, the validity for the given variable will be returned.

        Parameters
        ----------
        name : str, optional
            Variable name; default None = All variables will be tested

        Returns
        -------
        Dict[str, Validity] or Validity
            (Dictionary of) the variable(s) value validity
        """
      
        if name is None:
            return dict(
                (name, variable.is_valid())
                for name, variable in self._variables.items()
            )
        else:           
            return self._variables[name].is_valid()

    def get_validity_ground(self,
        status: Validity,
        name: Optional[str] = None,
    ) -> Union[Dict[str, str], str]:
        # TODO unit tests
        """Get the ground arguments used to established the variable validity range.

        The status `Validity.OK` has no ground arguments.

        Parameters
        ----------
        status : Validity.{OK, WARNING, ERROR}
            Validity status for which the reasons are looked for.
        name : str, optional
            Variable name; default None = All variables will be tested

        Returns
        -------
        Dict[str, str] or str
            (Dictionary of) the variable(s) validity ground
        """
        if name is None:
            return dict(
                (name, variable.get_validity_comment(status))
                for name, variable in self._variables.items()
            )
        else:
            return self._variables[name].get_validity_comment(status)

    def copy(self,
        name: Optional[str] = None,
        direction: Optional[PortType] = None,
    ) -> "ExtensiblePort":
        """Duplicates the port.

        Parameters
        ----------
        name : str, optional
            Name of the duplicated port; default original port name
        direction : {PortType.IN, PortType.OUT}, optional
            Direction of the duplicated port; default original direction

        Returns
        -------
        Port
            The copy of the current port
        """
        new_name = self.name if name is None else name
        new_direction = self.direction if direction is None else direction

        new_port = type(self)(new_name, new_direction)
        for name, variable in self._variables.items():
            # Validation criteria are removed to avoid warning duplication when checking
            new_port.add_variable(
                name,
                getattr(self, name),
                unit=variable.unit,
                dtype=variable.dtype,
                desc=variable.description,
                distribution=variable.distribution,
                scope=variable.scope,
            )

        return new_port

    def morph(self, port: "ExtensiblePort") -> None:
        """Morph the provided port into this port.

        Morphing a port is useful when converting a `System` in something equivalent. The morphing
        goal is to preserve connections and adapt the port content if needed.

        Parameters
        ----------
        port : ExtensiblePort
            The port to morph to
        """

        # TODO unit test for variable details when morphing
        for name, variable in self._variables.items():
            if name not in port:
                port.add_variable(
                    name,
                    getattr(self, name),
                    unit=variable.unit,
                    dtype=variable.dtype,
                    desc=variable.description,
                    valid_range=variable.valid_range,
                    invalid_comment=variable.invalid_comment,
                    limits=variable.limits,
                    out_of_limits_comment=variable.out_of_limits_comment,
                    distribution=variable.distribution,
                    scope=variable.scope,
                )

        for variable in list(port):
            if variable not in self:
                port.remove_variable(variable)

    def to_dict(self, with_def: bool = False) -> Dict[str, Union[str, Tuple[Dict[str, str], str]]]:
        """Convert this port in a dictionary.
   
        Parameters
        ----------
        with_def : bool
            Flag to export also output ports and its class name (default: False).

        Returns
        -------
        dict
            The dictionary representing this port.
        """
        # TODO this is uncomplete as validation ranges and distribution could be changed
        new_dict = dict()

        if with_def:
            tmp =  dict()
            if self.name not in ["inwards", "outwards"]:
                tmp["__class__"] = self.__class__.__qualname__
                for variable in self:
                    tmp[variable] = getattr(self, variable)
            else:
                for v_name, variable in self._variables.items():
                    tmp[v_name]= variable.to_dict() 

            new_dict[self.name] = tmp
        
        else:
            if self.direction == PortType.IN:
                for variable in self:
                    fullname = f"{self.name}.{variable}"
                    new_dict[fullname] = getattr(self, variable)
        
        return new_dict


class Port(ExtensiblePort):
    """A `Port` is a container gathering variables tightly linked.

    An optional dictionary may be specified to overwrite variable value and some of their
    metadata. Values of the dictionary may be of two kinds:
    - Only the new value
    - A dictionary with new value and/or new details

    If the value or a metadata is set to `None`, the default value will be kept.

    Parameters
    ----------
    name : str
        `Port` name
    direction : {PortType.IN, PortType.OUT}
        `Port` direction
    variables : Dict[str, Any], optional
        Dictionary of variables with their value and details; default: None = default value

    Attributes
    ----------
    _locked : bool
        if True, `add_variable` is deactivated. This is the default behavior outside the `setup`
        function.

    Examples
    --------

    Use this class by subclassing it:

    >>> class FlowPort(Port):
    >>>     def setup(self):
    >>>         self.add_variable('Pt', 101325., unit='Pa')
    >>>         self.add_variable('Tt', 273.15, unit='K')
    >>>         self.add_variable('W', 1.0, unit='kg/s')
    >>>
    >>> p = FlowPort('myPort', PortType.IN)
    >>> # Overwrite value and some details
    >>> f = FlowPort(
    >>>     'myPort2', 
    >>>     PortType.OUT,
    >>>     {
    >>>         'Pt': 10e6,
    >>>         'W': {
    >>>             'value': 10.,  # New value
    >>>             'unit': 'lbm/s',  # New unit - should be compatible of the original one
    >>>             'valid_range': (0., None),  # Updated validated range
    >>>             'invalid_comment': 'Flow should be positive',  # Comment if out of validated range
    >>>             'limits': None,  # Limits
    >>>             'out_of_limits_comment': '',  # Comment if out of limits
    >>>         }
    >>>     }
    >>> )
    """

    def __init__(
        self, name: str, direction: PortType, variables: Optional[Dict[str, Any]] = None
    ) -> None:
        """`Port` constructor.

        An optional dictionary may be specified to overwrite variable value and some of their
        metadata. Values of the dictionary may be of two kinds:
        - Only the new value
        - A dictionary with new value and/or new details

        Parameters
        ----------
        name : str
            `Port` name
        direction : {PortType.IN, PortType.OUT}
            `Port` direction
        variables : Dict[str, Any], optional
            Dictionary of variables with their value and details; default: None = default value
        """
        check_arg(variables, "variables", (type(None), dict))

        super().__init__(name, direction)
        self._locked = False  # type: bool

        self.setup()

        if variables is not None:
            for name, value in variables.items():
                variable_value = value
                if isinstance(value, dict):
                    variable_value = value.pop("value", None)

                    details = self.get_details(name)
                    try:
                        for field, param in value.items():
                            setattr(details, field, param)
                    except AttributeError:  # Unexpected keyword for details
                        # If current variable is of type dict
                        if isinstance(self[name], dict):
                            if variable_value is not None:
                                value["value"] = variable_value
                            variable_value = value
                        else:
                            raise

                if variable_value is not None:
                    self[name] = variable_value

        self._locked = True

    def setup(self) -> None:
        """`Port` variables are defined in this function by calling `add_variable`.

        This function allows to populate a customized `Port` class. The `add_variable`
        function is only callable in the frame of this function.

        Examples
        --------

        Here is an example of `Port` subclassing:

        >>> class FlowPort(Port):
        >>>     def setup(self):
        >>>         self.add_variable('Pt', 101325.)
        >>>         self.add_variable('Tt', 273.15)
        >>>         self.add_variable('W', 1.0)
        """
        pass  # pragma: no cover

    def add_variable(
        self,
        name: str,
        value: Any = 1,
        unit: str = "",
        dtype: Types = None,
        valid_range: RangeValue = None,
        invalid_comment: str = "",
        limits: RangeValue = None,
        out_of_limits_comment: str = "",
        desc: str = "",
        distribution: Optional[Distribution] = None,
        scope=Scope.PUBLIC,
    ) -> None:
        """Add a variable to the port.

        The `valid_range` defines the range of value for which a model is known to behave correctly.
        The `limits` are at least as large as the validity range.

        Notes
        -----
        The default visibility of a `Port` variable is `Scope.PUBLIC`. This is a difference compared
        to `ExtensiblePort`.

        Parameters
        ----------
        name : str
            Name of the variable
        value : Any, optional
            Value of the variable; default 1
        unit : str, optional
            Variable unit; default empty string (i.e. dimensionless)
        dtype : type or iterable of type, optional
            Variable type; default None (i.e. type of initial value)
        valid_range : Tuple[Any, Any] or Tuple[Tuple], optional
            Validity range of the variable; default None (i.e. all values are valid)
        invalid_comment : str, optional
            Comment to show in case the value is not valid; default ''
        limits : Tuple[Any, Any] or Tuple[Tuple], optional
            Limits over which the use of the model is wrong; default valid_range
        out_of_limits_comment : str, optional
            Comment to show in case the value is not valid; default ''
        desc : str, optional
            Variable description; default ''        
        distribution : Distribution, optional
            Probability distribution of the variable; default None (no distribution)
        scope : Scope {PRIVATE, PROTECTED, PUBLIC}, optional
            Variable visibility; default PUBLIC
        """
        if self._locked:
            raise AttributeError("add_variable cannot be called outside `setup`.")

        super().add_variable(
            name,
            value=value,
            unit=unit,
            dtype=dtype,
            valid_range=valid_range,
            invalid_comment=invalid_comment,
            limits=limits,
            out_of_limits_comment=out_of_limits_comment,
            desc=desc,
            distribution=distribution,
            scope=scope,
        )

    # TODO remove ?
    def remove_variable(self, name: str) -> None:
        """Remove the variable from the port.

        Parameters
        ----------
        name : str
            Name of the variable to be removed

        Raises
        ------
        ValueError
            If the variable does not exists
        """
        raise AttributeError("remove_variable cannot be called on Port.")

    def copy(
        self, name: Optional[str] = None, direction: Optional[PortType] = None
    ) -> "ExtensiblePort":
        """Duplicates the port.

        Parameters
        ----------
        name : str, optional
            Name of the duplicated port; default original port name
        direction : {PortType.IN, PortType.OUT}, optional
            Direction of the duplicated port; default original direction

        Returns
        -------
        Port
            The copy of the current port
        """
        new_name = self.name if name is None else name
        new_direction = self.direction if direction is None else direction

        new_port = type(self)(new_name, new_direction)

        for name in self:
            new_port[name] = self[name]

        return new_port



import pytest

import numpy as np
from cosapp.drivers.time.utils import TimeUnknownStack, TimeVarManager
from cosapp.systems import System
from .conftest import PointMass, PointMassWithPorts, DevilCase


@pytest.fixture(scope="function")
def make_case():
    def factory(cls, name, **options):
        system = cls(name, **options)
        manager = TimeVarManager(system)
        return system, manager
    return factory


@pytest.mark.parametrize("cls, branch, name", [
    (PointMass, ['x', 'v', 'a'], "[x, v]"),
    (PointMassWithPorts, ['position.x', 'kinematics.v', 'a'], "[position.x, kinematics.v]"),
])
def test_TimeUnknownStack__init__(make_case, cls, branch, name):
    point, manager = make_case(cls, "point")
    assert set(manager.transients.keys()) == {name}
    stack = manager.transients[name]
    assert isinstance(stack, TimeUnknownStack)
    assert stack.name == name


def test_TimeUnknownStack_der_1(make_case):
    point, manager = make_case(PointMass, "point")
    stack = manager.transients["[x, v]"]
    assert np.array_equal(stack.value, np.zeros(6))

    point.x = np.zeros_like(point.x)
    point.v = np.full_like(point.v, 0.1)
    point.k = 0.0
    point.run_once()
    # stack.value is not changed by x or v, because the object holds
    # a copy of [x, v], evaluated at construction. Should this be changed?
    assert np.array_equal(stack.value, np.zeros(6))
    # time derivative and max dt are dynamically evaluated
    assert np.array_equal(stack.d_dt, [0.1, 0.1, 0.1, 0, 0, -9.81])
    assert stack.max_time_step == pytest.approx(0.981)

    point.k = 0.1
    point.mass = 1.0
    point.run_once()
    assert np.array_equal(stack.value, np.zeros(6))
    assert np.allclose(point.a, [-0.01, -0.01, -9.82], rtol=1e-12)
    assert np.allclose(stack.d_dt, [0.1, 0.1, 0.1, -0.01, -0.01, -9.82], rtol=1e-12)
    assert stack.max_time_step == pytest.approx(0.1 * np.linalg.norm(point.a))


def test_TimeUnknownStack_der_2(make_case):
    point, manager = make_case(PointMassWithPorts, "point")
    stack = manager.transients["[position.x, kinematics.v]"]
    assert np.array_equal(stack.value, np.zeros(6))

    point.position.x = np.zeros_like(point.position.x)
    point.kinematics.v = np.full_like(point.kinematics.v, 0.1)
    point.k = 0.0
    point.run_once()
    # stack.value is not changed by x or v, because the object holds
    # a copy of [x, v], evaluated at construction. Should this be changed?
    assert np.array_equal(stack.value, np.zeros(6))
    # time derivative and max dt are dynamically evaluated
    assert np.array_equal(stack.d_dt, [0.1, 0.1, 0.1, 0, 0, -9.81])
    assert stack.max_time_step == pytest.approx(0.981)

    point.k = 0.1
    point.mass = 1.0
    point.run_once()
    a = point.a
    assert np.array_equal(stack.value, np.zeros(6))
    assert np.allclose(a, [-0.01, -0.01, -9.82], rtol=1e-12)
    assert np.allclose(stack.d_dt, [0.1, 0.1, 0.1, -0.01, -0.01, -9.82], rtol=1e-12)
    assert stack.max_time_step == pytest.approx(0.1 * np.linalg.norm(a))

@pytest.mark.parametrize("pulling", [
    None, ["x"], ["x", "v"], ["v", "a"], ["a"], ["alpha"], ["omega", "zeta"], ["zeta"]
])
def test_TimeUnknownStack_multisystems(pulling):
    devil = DevilCase("devil")
    top = System("top")
    top.add_child(devil, pulling=pulling)
    manager = TimeVarManager(top)
    assert len(manager.transients) == 4
    stack_vector = manager.transients[f"{devil.name}[x, v]"]
    assert np.array_equal(stack_vector.value, np.zeros(6))

    stack_scalar = manager.transients[f"{devil.name}[alpha, omega]"]
    assert np.array_equal(stack_scalar.value, np.zeros(2))

    new_vector =  np.random.randn(6)
    stack_vector.value = new_vector
    new_scalar = np.random.randn(2)
    stack_scalar.value = new_scalar
    top.run_once()  # Propagate the unknown if a pulling occurred
    assert devil.x == pytest.approx(new_vector[:3])
    assert devil.v == pytest.approx(new_vector[3:])
    assert devil.alpha == pytest.approx(new_scalar[0])
    assert devil.omega == pytest.approx(new_scalar[1])


def test_TimeUnknownStack_value(make_case):
    point, manager = make_case(PointMass, "point")
    stack = manager.transients["[x, v]"]
    # Test value getter
    assert np.array_equal(stack.value[:3], point.x)
    assert np.array_equal(stack.value[3:], point.v)
    # Test value setter
    stack.value = np.array(range(6), dtype=float)
    assert np.array_equal(point.x, [0, 1, 2])
    assert np.array_equal(point.v, [3, 4, 5])



from collections.abc import MutableMapping
from numbers import Number
from typing import Any, Dict, Iterator, List, Optional, Tuple, TypeVar, Union

import numpy

from cosapp.core.eval_str import EvalString
from cosapp.core.variableref import VariableReference
from cosapp.core.numerics.boundary import AbstractTimeUnknown, TimeUnknown
from cosapp.systems.system import System
from cosapp.utils.helpers import check_arg

T = TypeVar('T')


class TimeUnknownStack(AbstractTimeUnknown):
    """
    Class representing a group of variables [a, b, c, ...] jointly solved by a time driver, with
    b = da/dt, c = db/dt and so on. By nature, the unknown is therefore an array.
    If variables a, b, c... are arrays themselves, they are automatically flattened.

    Parameters
    ----------
    context : System
        System CoSApp in which all transients to be stacked are defined
    name : str
        Name of this time unknown stack
    transients : List[TimeUnknown]
        Stacked unknowns

    Notes
    -----
    The group variables must all be defined as variables of the same system.
    """
    def __init__(self,
        context: System,
        name: str,
        transients: List[TimeUnknown],
    ):
        super().__init__()
        self.__context = context
        self.__name = name
        self.__value = None
        self.__transients = transients
        self.__init_stack()

    @property
    def name(self) -> str:
        """str: Name of the variable"""
        return self.__name

    @property
    def context(self) -> System:
        """System: Evaluation context of the stacked unknown"""
        return self.__context

    @property
    def der(self) -> EvalString:
        """Expression of time derivative of stacked vector, given as an EvalString"""
        return self.__der

    def __str__(self) -> str:
        return str(self.value)

    def __repr__(self) -> str:
        return f"{self.__name} := {self!s}"

    def __init_stack(self) -> None:
        """
        1. Update the expression of the time derivative vector, and
        the size of individual variables (1 for scalars, n > 1 for vectors).
        It is assumed that all variables in the stack have the same size; in principle,
        this property is satisfied by design, since size consistency between a
        variable and its derivative is checked at declaration of all transient variables.
        
        2. Update the expression of max_time_step, defined as the minimum value of all
        individual maximum time steps (if any) defined by the stacked variables.
        """
        root = self.__transients[0].value

        size = len(root) if isinstance(root, (list, numpy.ndarray)) else 1
        # Init self.__value
        self.reset()

        expr = "asarray([{}]).ravel()"
        get_stack = lambda args: EvalString(expr.format(", ".join(args)), self.__context)
        self.__der = get_stack(map(lambda u: str(u.der), self.__transients))
        self.__var_size = size
        if size == 1:
            self.__sub_value = lambda offset: self.__value[offset]
        else:
            self.__sub_value = lambda offset: self.__value[offset : offset + size]

        # Set expressions of max time step and max absolute step
        def step_expr(mapping):
            args = ", ".join(map(mapping, self.__transients))
            return EvalString(f"min({args})", self.__context)
        self.__max_dt = step_expr(lambda var: str(var.max_time_step_expr))
        self.__max_dx = step_expr(lambda var: str(var.max_abs_step_expr))

    @property
    def value(self) -> numpy.ndarray:
        """numpy.ndarray: Value of the time unknown"""
        return self.__value

    @value.setter
    def value(self, new: Union[List[float], numpy.ndarray]) -> None:
        if numpy.shape(new) != self.__value.shape:
            raise ValueError("Incompatible array shapes")
        self.__value = numpy.array(new)
        # Update original system variables
        size = self.__var_size
        for i, unknown in enumerate(self.__transients):
            offset = i * size
            unknown.value = self.__sub_value(offset)

    def reset(self) -> None:
        """Reset stack value from original system variables"""
        self.__value = numpy.array(list(map(lambda t: t.value, self.__transients))).ravel()

    @property
    def max_time_step_expr(self) -> EvalString:
        """EvalString: Expression of the maximum time step allowed for the instantaneous time evolution of the stacked variable"""
        return self.__max_dt

    @property
    def max_abs_step_expr(self) -> EvalString:
        """EvalString: Expression of the maximum step allowed for the stacked variable"""
        return self.__max_dx


class TimeUnknownDict(MutableMapping):
    """
    Dictionary of AbstractTimeUnknown objects, mapped to str variable names.
    Automatically updates a dictionary of time step constrained variables,
    accessible with read-only property `constrained`.
    """
    def __init__(self, **mapping):
        super().__init__()
        self.__transients = {}
        self.__constrained = {}
        self.update(mapping)

    def __str__(self) -> str:
        return str(self.__transients)

    def __repr__(self) -> str:
        return repr(self.__transients)

    def __len__(self) -> int:
        """int: length of the collection"""
        return len(self.__transients)

    def __getitem__(self, key: str) -> AbstractTimeUnknown:
        return self.__transients[key]

    def __setitem__(self, key: str, value: AbstractTimeUnknown) -> None:
        if not isinstance(key, str):
            raise TypeError(
                f"Keys of TimeUnknownDict must be strings; invalid key {key!r}"
            )
        if not isinstance(value, AbstractTimeUnknown):
            raise TypeError(
                "Elements of TimeUnknownDict must be of type AbstractTimeUnknown"
                f"; invalid item {key!r}: {type(value).__qualname__}"
            )
        self.__transients[key] = value
        if value.constrained:
            self.__constrained[key] = value
        else:
            try:
                self.__constrained.pop(key)
            except KeyError:
                pass

    def __delitem__(self, key: str) -> None:
        self.__transients.__delitem__(key)
        try:
            self.__constrained.__delitem__(key)
        except KeyError:
            pass

    def __contains__(self, key: str) -> bool:
        return key in self.__transients

    def __iter__(self) -> Iterator[str]:
        """Iterator on dictionary keys."""
        return iter(self.__transients)

    def __str__(self) -> str:
        return str(self.__transients)

    def __repr__(self) -> repr:
        return str(self.__transients)

    def keys(self, constrained=False) -> Iterator[str]:
        """Iterator on dictionary keys, akin to dict.keys().
        If `constrained` is True, the iterator applies only to time step constrained variables."""
        return self.__constrained.keys() if constrained else self.__transients.keys()

    def values(self, constrained=False) -> Iterator[AbstractTimeUnknown]:
        """Iterator on dictionary values, akin to dict.values().
        If `constrained` is True, the iterator applies only to time step constrained variables."""
        return self.__constrained.values() if constrained else self.__transients.values()

    def items(self, constrained=False) -> Iterator[Tuple[str, AbstractTimeUnknown]]:
        """Iterator on (key, value) tuples, akin to dict.items().
        If `constrained` is True, the iterator applies only to time step constrained variables."""
        return self.__constrained.items() if constrained else self.__transients.items()

    def get(self, key: str, *default: Optional[Any]) -> AbstractTimeUnknown:
        """Get value associated to `key`. Behaves as dict.get()."""
        return self.__transients.get(key, *default)

    def pop(self, key: str, *default: Optional[Any]) -> AbstractTimeUnknown:
        """Pop value associated to `key`. Behaves as dict.pop()."""
        try:
            self.__constrained.pop(key)
        except KeyError:
            pass
        finally:
            return self.__transients.pop(key, *default)

    def update(self, mapping: Dict) -> None:
        for key, value in mapping.items():
            self.__setitem__(key, value)

    def clear(self) -> None:
        self.__transients.clear()
        self.__constrained.clear()

    def max_time_step(self) -> float:
        """
        Compute maximum admissible time step, from transients' max_time_step (inf by default).
        Raises ValueError if any transients' max_time_step is not strictly positive.
        """
        dt = numpy.inf
        for name, transient in self.__constrained.items():
            max_dt = transient.max_time_step
            if max_dt <= 0:
                raise RuntimeError(
                    f"The maximum time step of {name} was evaluated to non-positive value {max_dt}"
                )
            dt = min(dt, max_dt)
        return dt

    @property
    def constrained(self) -> Dict[str, AbstractTimeUnknown]:
        """Dict[str, AbstractTimeUnknown]: shallow copy of the subset of time step constrained variables."""
        return self.__constrained.copy()


class TimeVarManager:
    """
    Class dedicated to the analysis of a system's independent transient variables.
    For example, in a system where

    .. math::

       dH/dt = f(a, b),

       dx/dt = v,

       dv/dt = a,

    the manager will identify two independent variables H and [x, v], with time derivatives
    f(a, b) and [v, a], respectively. Variable [x, v] and its derivative are handled by class
    `TimeUnknownStack`.
    """
    def __init__(self, context: System):
        self.context = context

    @property
    def context(self) -> System:
        """System handled by manager"""
        return self.__context

    @property
    def problem(self) -> "MathematicalProblem":
        """Mathematical problem handled by manager"""
        return self.__problem

    @context.setter
    def context(self, context: System):
        if not isinstance(context, System):
            raise TypeError("TimeVarManager context must be a system")
        self.__context = context
        self.update_transients()

    @property
    def transients(self) -> TimeUnknownDict:
        """
        Dictionary of all transient variables in current system, linking each
        variable (key) to its associated time unknown (value).
        For stand-alone, order-1 derivatives, transient unknowns are of type `TimeUnknown`.
        For higher-order derivatives, related variables are gathered into a `TimeUnknownStack` object.
        """
        return self.__transients

    @property
    def rates(self) -> Dict[str, "TimeDerivative"]:
        """
        Dictionary of all rate variables in current system, linking each
        variable (key) to its associated TimeDerivative object (value).
        """
        return self.__problem.rates

    def update_transients(self) -> None:
        """Update the transient variable dictionary (see property `transients` for details)"""
        context = self.__context
        problem = context.get_unsolved_problem()
        context_transients = problem.transients
        ders = dict()
        reference2name = dict()
        for name, unknown in context_transients.items():
            reference = unknown.pulled_from or unknown.context.name2variable[unknown.name]
            reference2name[reference] = name
            der_context = unknown.der.eval_context
            derivative_expr = str(unknown.der)
            try:
                ders[reference] = der_context.name2variable[derivative_expr]
            except KeyError:   # Complex derivative expression
                ders[reference] = VariableReference(context=der_context, mapping=None, key=derivative_expr)

        transients = TimeUnknownDict()
        # Comparison is done with VariableReference as
        # syst.name2variable["phi"] == syst.name2variable["inwards.phi"]
        tree = self.get_tree(ders)
        for root, branch in tree.items():
            if len(branch) > 2:  # second- or higher-order derivative -> build unknown stack
                stack_context = root.context
                branches = list(map(TimeVarManager._get_variable_fullname, branch[:-1]))
                root_stack_name = ", ".join(branches).join("[]")
                context_name = context.get_path_to_child(stack_context)
                stack_name = f"{context_name}{root_stack_name}"
                transients_stack = map(
                    lambda name: context_transients[name], 
                    map(lambda reference: reference2name[reference], branch[:-1])
                )
                transients[stack_name] = TimeUnknownStack(
                    stack_context, stack_name, list(transients_stack))
            else:  # first-order time derivative -> use current unknown
                root_name = reference2name[root]
                transients[root_name] = context_transients[root_name]

        self.__transients = transients
        self.__problem = problem

    @staticmethod
    def _get_variable_fullname(ref: VariableReference) -> str:
        """Built the variable fullname for its reference.
        
        Parameters
        ----------
        ref : VariableReference
            Reference to the variable
        
        Returns
        -------
        str
            The variable fullname
        """
        # First condition is to handle complex derivative expression see previous loop
        if ref.mapping is None or ref.mapping.name in (System.INWARDS, System.OUTWARDS):
            return ref.key
        else:
            return f"{ref.mapping.name}.{ref.key}"

    @staticmethod
    def get_tree(ders: Dict[T, T]) -> Dict[T, List[T]]:
        """
        Parse a dictionary of the kind (var, d(var)/dt), to detect a dependency
        chain from one root variable to its successive time derivatives.
        Returns a dictionary of the kind:
        (root var 'X', [X_0, X_1, .., X_n]), where X_n is the expression of
        the nth-order time derivative of X.
        """
        var_list = ders.keys()
        der_list = ders.values()
        roots = list(filter(lambda var: var not in der_list, var_list))
        leaves = list(filter(lambda der: der not in var_list, der_list))
        tree = dict()
        for root in roots:
            tree[root] = [root]
            var = root
            while ders[var] not in leaves:
                der = ders[var]
                tree[root].append(der)
                var = der
            tree[root].append(ders[var])
        return tree

    def max_time_step(self) -> float:
        """
        Compute maximum admissible time step, from transients' `max_time_step` (numpy.inf by default).
        Raises ValueError if any transients' max_time_step is not strictly positive.
        """
        return self.__transients.max_time_step()


class TimeStepManager:
    """
    Class dedicated to the management of time step for time drivers.
    """
    def __init__(self, transients=TimeUnknownDict(), dt=None, max_growth_rate=None):
        self.__dt = None
        self.__nominal_dt = None
        self.__transients = None
        self.__growthrate = None
        # Assign initial values
        self.transients = transients
        self.nominal_dt = dt
        self.max_growth_rate = max_growth_rate

    @property
    def transients(self):
        return self.__transients

    @transients.setter
    def transients(self, transients: TimeUnknownDict):
        check_arg(transients, "transients", TimeUnknownDict)
        self.__transients = transients

    @property
    def nominal_dt(self) -> Number:
        """Time step"""
        return self.__nominal_dt

    @nominal_dt.setter
    def nominal_dt(self, value: Number) -> None:
        if value is not None:
            check_arg(value, 'dt', Number, value_ok = lambda dt: dt > 0)
        self.__nominal_dt = value

    @property
    def max_growth_rate(self) -> Number:
        """Maximum growth rate of time step"""
        return self.__growthrate

    @max_growth_rate.setter
    def max_growth_rate(self, value: Number) -> None:
        if value is None:
            self.__growthrate = numpy.inf
        else:
            check_arg(value, 'max_growth_rate', Number, value_ok = lambda x: x > 1)
            self.__growthrate = value

    def max_time_step(self) -> float:
        """
        Compute maximum admissible time step, from transients' max_time_step (inf by default).
        Raises ValueError if any transients' max_time_step is not strictly positive.
        """
        return self.__transients.max_time_step()

    def time_step(self, previous=None) -> float:
        """
        Compute time step, making sure that it does not exceed any transient's `max_time_step`
        (numpy.inf by default), and that all transient max_time_step are strictly positive.
        If `previous` is specified, the returned time step is bounded by max_growth_rate * previous.

        An exception is raised if time step is ultimately found to be infinity.
        """
        dt = self.__nominal_dt or numpy.inf
        dt = min(dt, self.max_time_step())
        if previous is not None and previous > 0:
            dt = min(dt, self.__growthrate * previous)

        if not numpy.isfinite(dt):
            raise ValueError("Time step was not specified, and could not be determined from transient variables")

        return dt
